# coding: utf-8

"""
    

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)

    OpenAPI spec version: beta
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


from pprint import pformat

from six import iteritems


class DataSource(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """


    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'name': 'str',
        'format': 'str',
        'categories': 'list[str]',
        'pk': 'DataSourcePK'
    }

    attribute_map = {
        'name': 'name',
        'format': 'format',
        'categories': 'categories',
        'pk': 'pk'
    }

    def __init__(self, name=None, format=None, categories=None, pk=None):
        """
        DataSource - a model defined in Swagger
        """

        self._name = None
        self._format = None
        self._categories = None
        self._pk = None

        self.name = name
        if format is not None:
          self.format = format
        if categories is not None:
          self.categories = categories
        self.pk = pk

    @property
    def name(self):
        """
        Gets the name of this DataSource.

        :return: The name of this DataSource.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """
        Sets the name of this DataSource.

        :param name: The name of this DataSource.
        :type: str
        """
        if name is None:
            raise ValueError("Invalid value for `name`, must not be `None`")

        self._name = name

    @property
    def format(self):
        """
        Gets the format of this DataSource.
        datasource format (a file type or even a data-store type), e.g. parquet, csv, mysql, kafka

        :return: The format of this DataSource.
        :rtype: str
        """
        return self._format

    @format.setter
    def format(self, format):
        """
        Sets the format of this DataSource.
        datasource format (a file type or even a data-store type), e.g. parquet, csv, mysql, kafka

        :param format: The format of this DataSource.
        :type: str
        """

        self._format = format

    @property
    def categories(self):
        """
        Gets the categories of this DataSource.
        datasource categories (Their IDs)

        :return: The categories of this DataSource.
        :rtype: list[str]
        """
        return self._categories

    @categories.setter
    def categories(self, categories):
        """
        Sets the categories of this DataSource.
        datasource categories (Their IDs)

        :param categories: The categories of this DataSource.
        :type: list[str]
        """

        self._categories = categories

    @property
    def pk(self):
        """
        Gets the pk of this DataSource.

        :return: The pk of this DataSource.
        :rtype: DataSourcePK
        """
        return self._pk

    @pk.setter
    def pk(self, pk):
        """
        Sets the pk of this DataSource.

        :param pk: The pk of this DataSource.
        :type: DataSourcePK
        """
        if pk is None:
            raise ValueError("Invalid value for `pk`, must not be `None`")

        self._pk = pk

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        if not isinstance(other, DataSource):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



import pickle as pk
from kensu.utils.kensu_provider import KensuProvider
from kensu.utils.helpers import eventually_report_in_mem


def wrap_dump(method):
    def wrapper(*args, **kwargs):

        if ('kensu.sklearn') in str(args[0].__class__):
            fmt = 'Sklearn Model'

        else:
            fmt = 'Unknown'
        kensu = KensuProvider().instance()

        result = method(*args, **kwargs)

        loc = args[1].name
        model = args[0]

        def get_absolute_path(path):
            import os
            return 'file:'+str(os.path.abspath(path))

        location = get_absolute_path(loc)



        orig_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(model.attr[0], kensu.default_physical_location_ref,
                                                     logical_naming=kensu.logical_naming,format=fmt))
        orig_sc = eventually_report_in_mem(kensu.extractors.extract_schema(orig_ds, model.attr[0]))


        result_ds = kensu.extractors.extract_data_source(model.attr[0], kensu.default_physical_location_ref, location = location,
                                                       logical_naming=kensu.logical_naming,format=fmt)._report()
        result_sc = kensu.extractors.extract_schema(result_ds, model.attr[0])._report()

        kensu.real_schema_df[result_sc.to_guid()] = model
        kensu.model[result_sc.to_guid()] = model.attr

        if kensu.mapping == True:
            for col in [s.name for s in result_sc.pk.fields]:
                kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), orig_sc.to_guid(), str(col),
                                             "Pickle Dump")
            kensu.report_with_mapping()


        return result

    wrapper.__doc__ = method.__doc__
    return wrapper

def wrap_load(method):
    def wrapper(*args, **kwargs):
        kensu = KensuProvider().instance()
        result = method(*args, **kwargs)

        loc = args[0].name
        model = result

        def get_absolute_path(path):
            import os
            return 'file:'+str(os.path.abspath(path))

        location = get_absolute_path(loc)

        if ('sklearn') in str(args[0].__class__):
            fmt = 'Sklearn Model'
        else:
            fmt = 'Unknown'

        orig_ds = (kensu.extractors.extract_data_source(model.attr[0], kensu.default_physical_location_ref,
                                                                              location=location, logical_naming=kensu.logical_naming,format=fmt))._report()
        orig_sc = (kensu.extractors.extract_schema(orig_ds, model.attr[0]))._report()


        result_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(model.attr[0], kensu.default_physical_location_ref,
                                                       logical_naming=kensu.logical_naming,format=fmt))
        result_sc = eventually_report_in_mem(kensu.extractors.extract_schema(result_ds, model.attr[0]))

        kensu.real_schema_df[orig_sc.to_guid()] = result
        kensu.model[orig_sc.to_guid()] = model.attr

        if kensu.mapping == True:
            for col in [s.name for s in result_sc.pk.fields]:
                kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), orig_sc.to_guid(), str(col),
                                             "Pickle Load")


        return result

    wrapper.__doc__ = method.__doc__
    return wrapper


# coding: utf-8

"""
    

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)

    OpenAPI spec version: beta
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


from pprint import pformat

from six import iteritems


class ProcessRun(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """


    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'pk': 'ProcessRunPK',
        'launched_by_user_ref': 'UserRef',
        'executed_code_version_ref': 'CodeVersionRef',
        'environment': 'str',
        'projects_refs': 'list[ProjectRef]'
    }

    attribute_map = {
        'pk': 'pk',
        'launched_by_user_ref': 'launchedByUserRef',
        'executed_code_version_ref': 'executedCodeVersionRef',
        'environment': 'environment',
        'projects_refs': 'projectsRefs'
    }

    def __init__(self, pk=None, launched_by_user_ref=None, executed_code_version_ref=None, environment=None, projects_refs=None):
        """
        ProcessRun - a model defined in Swagger
        """

        self._pk = None
        self._launched_by_user_ref = None
        self._executed_code_version_ref = None
        self._environment = None
        self._projects_refs = None

        self.pk = pk
        if launched_by_user_ref is not None:
          self.launched_by_user_ref = launched_by_user_ref
        if executed_code_version_ref is not None:
          self.executed_code_version_ref = executed_code_version_ref
        if environment is not None:
          self.environment = environment
        if projects_refs is not None:
          self.projects_refs = projects_refs

    @property
    def pk(self):
        """
        Gets the pk of this ProcessRun.

        :return: The pk of this ProcessRun.
        :rtype: ProcessRunPK
        """
        return self._pk

    @pk.setter
    def pk(self, pk):
        """
        Sets the pk of this ProcessRun.

        :param pk: The pk of this ProcessRun.
        :type: ProcessRunPK
        """
        if pk is None:
            raise ValueError("Invalid value for `pk`, must not be `None`")

        self._pk = pk

    @property
    def launched_by_user_ref(self):
        """
        Gets the launched_by_user_ref of this ProcessRun.
        a user responsible for the start of the process

        :return: The launched_by_user_ref of this ProcessRun.
        :rtype: UserRef
        """
        return self._launched_by_user_ref

    @launched_by_user_ref.setter
    def launched_by_user_ref(self, launched_by_user_ref):
        """
        Sets the launched_by_user_ref of this ProcessRun.
        a user responsible for the start of the process

        :param launched_by_user_ref: The launched_by_user_ref of this ProcessRun.
        :type: UserRef
        """

        self._launched_by_user_ref = launched_by_user_ref

    @property
    def executed_code_version_ref(self):
        """
        Gets the executed_code_version_ref of this ProcessRun.
        indicates which code version is executed in the given process

        :return: The executed_code_version_ref of this ProcessRun.
        :rtype: CodeVersionRef
        """
        return self._executed_code_version_ref

    @executed_code_version_ref.setter
    def executed_code_version_ref(self, executed_code_version_ref):
        """
        Sets the executed_code_version_ref of this ProcessRun.
        indicates which code version is executed in the given process

        :param executed_code_version_ref: The executed_code_version_ref of this ProcessRun.
        :type: CodeVersionRef
        """

        self._executed_code_version_ref = executed_code_version_ref

    @property
    def environment(self):
        """
        Gets the environment of this ProcessRun.
        indicates environment (Prod, Dev, Test) in which the process was executed

        :return: The environment of this ProcessRun.
        :rtype: str
        """
        return self._environment

    @environment.setter
    def environment(self, environment):
        """
        Sets the environment of this ProcessRun.
        indicates environment (Prod, Dev, Test) in which the process was executed

        :param environment: The environment of this ProcessRun.
        :type: str
        """

        self._environment = environment

    @property
    def projects_refs(self):
        """
        Gets the projects_refs of this ProcessRun.
        indicates the top-level Projects to which this process is assigned

        :return: The projects_refs of this ProcessRun.
        :rtype: list[ProjectRef]
        """
        return self._projects_refs

    @projects_refs.setter
    def projects_refs(self, projects_refs):
        """
        Sets the projects_refs of this ProcessRun.
        indicates the top-level Projects to which this process is assigned

        :param projects_refs: The projects_refs of this ProcessRun.
        :type: list[ProjectRef]
        """

        self._projects_refs = projects_refs

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        if not isinstance(other, ProcessRun):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



class Strategy(object):

    @staticmethod
    def from_dict(d):
        """
        Create a Strategy based on a dict where
        - Keys are the output field names
        - Values an array of the inputs fields names used to create the output
        :param d: a dict of output -> [input]
        :return a Strategy using the dict to build the mappings
        """
        return Strategy(lambda i, o: o in d.keys() and i in d[o])

    def __init__(self, mapping_function):
        self.mapping_function = mapping_function

    def map(self, add_deps_builder_element, type="data"):
        c = add_deps_builder_element.builder.control
        if type == "data":
            c = add_deps_builder_element.builder.data

        for o in add_deps_builder_element.builder.output_fields_name:
            for i in add_deps_builder_element.builder.input_fields_name:
                if self.mapping_function(i, o):
                    if o not in c:
                        c[o] = []
                    c[o].append(i)

        return add_deps_builder_element

    def or_else(self, next_strategy):
        """
        Chains current Strategy with a next one in case an input and output aren't mapping in the current one.
        This allows to define general strategies followed by more specific ones,
        like DIRECT.or_else(map_2_specific_fields_only_strategy)
        :param next_strategy: the next Strategy to be used if the current doesn't pass
        :return: a Strategy combining the mapping of the current one and the next one
        """

        def f(i, o):
            return self.mapping_function(i, o) or next_strategy.mapping_function(i, o)

        s = Strategy(f)
        return s


FULL = Strategy(lambda i, o: True)
DIRECT = Strategy(lambda i, o: i == o)
def from_simple_io_dict(di):
    return Strategy(lambda i, o: di[i]==o)
OUT_STARTS_WITH_IN = Strategy(lambda i,o: o[0:len(i)] == i)




import logging
import re
from typing import Optional

import pandas as pd
from pandas._typing import Axes, Dtype
from pandas.core.indexing import _iLocIndexer, _LocIndexer
from pandas.core.strings import StringMethods
from pandas.core.arrays.categorical import CategoricalAccessor

import numpy

from kensu.numpy import ndarray
from kensu.pandas.extractor import KensuPandasSupport
from kensu.utils.dsl.extractors.external_lineage_dtos import GenericComputedInMemDs
from kensu.utils.kensu_class_handlers import KensuClassHandlers
from kensu.utils.kensu_provider import KensuProvider
from kensu.utils.dsl import mapping_strategies
from kensu.utils.helpers import eventually_report_in_mem, extract_ksu_ds_schema, get_absolute_path

from kensu.client import *
from kensu.utils.wrappers import remove_ksu_wrappers


class KensuPandasDelegator(object):
    SKIP_KENSU_FIELDS = ["_DataFrame__k_df", "INTERCEPTORS"]
    SKIP_KENSU_METHODS = ["get_df", "kensu_init", "k_to_format", "to_string"]

    def __getitem__(self, key):
        ''' Item lookup: returns a Series with the values of the key `key`'''
        result = self.__getitem__(key)
        if isinstance(result, DataFrame) or isinstance(result, pd.DataFrame):
            # this case is handled by the callable property as [[col1, col2, ...]] is converted into `loc`
            return result

        # we assume the type is Series
        if not isinstance(result, pd.Series):
            raise RuntimeError("Type expected is Series, got " + str(type(result)))
        kensu = KensuProvider().instance()

        orig_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(self, kensu.default_physical_location_ref))
        orig_sc = eventually_report_in_mem(kensu.extractors.extract_schema(orig_ds, self))

        result_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(result, kensu.default_physical_location_ref))
        result_sc = eventually_report_in_mem(kensu.extractors.extract_schema(result_ds, result))

        kensu_s = Series.using(result)
        kensu_s.name = key

        if kensu.mapping:
                kensu.add_dependencies_mapping(result_sc.to_guid(),key,orig_sc.to_guid(),key,'__getitem__')

        else:
            kensu.add_dependency((self, orig_ds, orig_sc), (result, result_ds, result_sc),
                               mapping_strategy=mapping_strategies.DIRECT)
        return kensu_s



    def __getattribute__(self, name):
        if name == "__class__":
            attr_value = object.__getattribute__(self, name)
            return attr_value
        elif name in ['_mgr','_data','_flags']:
            attr_value = object.__getattribute__(self.get_df(), name)
            return attr_value
        else:
            attr_value = object.__getattribute__(self, name)

            if hasattr(attr_value, '__call__'):
                return KensuPandasDelegator.handle_callable(self, name, attr_value)

            else:
                return KensuPandasDelegator.handle_field(self, name, attr_value)


    @staticmethod
    def wrap_returned_df(kensu_df, returned, name, original):
        if isinstance(returned, pd.DataFrame) \
                and not isinstance(returned, DataFrame) \
                and original is not returned:

            logging.debug("Wrapping attr access resulting dataframe in KensuDF for property name=" + name)
            tmp = DataFrame.using(returned)
            returned = tmp
        elif isinstance(returned, numpy.ndarray):
            logging.debug("Wrapping attr access resulting numpy.ndarray in KensuDF for property name=" + name)
            tmp = ndarray.using(returned)
            returned = tmp
        return returned

    @staticmethod
    def handle_field(kensu_df, name, attr_value):
        # _DataFrame__k_df => points to DataFrame.__k_df
        if name in KensuPandasDelegator.SKIP_KENSU_FIELDS:
            # return Kensu DF attr value right away
            logging.debug("Kensu attribute name=" + name)
            return attr_value
        else:
            logging.debug("NOT Kensu, let's delegate attribute name=" + name)

            pd_df = kensu_df.get_df()
            df_attr = object.__getattribute__(pd_df, name)
            result = KensuPandasDelegator.wrap_returned_df(kensu_df, df_attr, name, pd_df)

            #adding the datasources, schemas and lineage to the dependencies
            if result is not None and isinstance(result, DataFrame):
                kensu = KensuProvider().instance()
                orig_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(pd_df, kensu.default_physical_location_ref))
                orig_sc = eventually_report_in_mem(kensu.extractors.extract_schema(orig_ds,pd_df))

                result_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(result, kensu.default_physical_location_ref))
                result_sc = eventually_report_in_mem(kensu.extractors.extract_schema(result_ds, result))

                kensu.add_dependency((pd_df, orig_ds, orig_sc), (result, result_ds, result_sc),
                                   mapping_strategy=mapping_strategies.OUT_STARTS_WITH_IN)
            elif result is not None and isinstance(result, ndarray):
                kensu = KensuProvider().instance()
                orig_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(pd_df, kensu.default_physical_location_ref))
                orig_sc = eventually_report_in_mem(kensu.extractors.extract_schema(orig_ds,pd_df))

                result_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(result, kensu.default_physical_location_ref))
                result_sc = eventually_report_in_mem(kensu.extractors.extract_schema(result_ds, result))

                kensu.add_dependency((pd_df, orig_ds, orig_sc), (result, result_ds, result_sc),
                                   mapping_strategy=mapping_strategies.OUT_STARTS_WITH_IN)


            return result

    @staticmethod
    def handle_callable(kensu_df, name, attr_value):
        if name not in KensuPandasDelegator.SKIP_KENSU_METHODS:
            pd_df = kensu_df.get_df()
            delegated_pd_attr = object.__getattribute__(pd_df, name)
            # for example, `iloc` is subscriptable and is a property, hence call this way: iloc[]
            if isinstance(getattr(type(pd_df), name), property):
                return KensuPandasDelegator.handle_callable_property(kensu_df, name, pd_df, delegated_pd_attr)
            else:
                docstring = delegated_pd_attr.__doc__
                return KensuPandasDelegator.create_function_wrapper(kensu_df, name, docstring)
        else:
            logging.debug("Kensu function name=" + name)

            # return Kensu DF function result right away
            def w(*args, **kwargs):
                return attr_value(*args, **kwargs)
            return w

    @staticmethod
    def create_function_wrapper(kensu_df, name, docstring):
        logging.debug("Kensu wrapping function name=" + name)

        def wrapper(*args, **kwargs):

            pd_df = kensu_df.get_df()
            df_attr = object.__getattribute__(pd_df, name)

            # call method on delegated df
            delegation_config = getattr(kensu_df, 'INTERCEPTORS', None)
            for regex_attr_name in delegation_config:
                result_regex = regex_attr_name.search(name)
                if name == 'to_records':
                    result_regex = None

                if result_regex is not None:
                    # if we pass the regex, there is an interception defined, so we execute it
                    name_of_intercept_method = delegation_config[regex_attr_name]
                    intercept = object.__getattribute__(kensu_df, name_of_intercept_method)
                    intercept(result_regex, *args, **kwargs)

            kensu = KensuProvider().instance()

            # set_axis updates the DataFrame directly (self) and returns None, we loose the initial dataframe
            # so we need to create the Kensu object before
            orig_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(pd_df, kensu.default_physical_location_ref))
            orig_sc = eventually_report_in_mem(kensu.extractors.extract_schema(orig_ds, pd_df))

            new_args=remove_ksu_wrappers(args)
            new_args = tuple(new_args)
            result = df_attr(*new_args, **kwargs)
            original_result = result
            result = KensuPandasDelegator.wrap_returned_df(kensu_df, result, name, pd_df)

            if result is None :
                #set_axis updates the DataFrame directly (self) and returns None, so we reset the result
                if name in ["_set_axis","_set_item"]:
                    result = kensu_df
            #adding the datasources, schemas and lineage to the dependencies
            if (result_regex is None and result is not None and (isinstance(result,DataFrame) or isinstance(result,ndarray))):

                result_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(result, kensu.default_physical_location_ref))
                result_sc = eventually_report_in_mem(kensu.extractors.extract_schema(result_ds, result))

                if kensu.mapping :
                    if name in ['__getitem__','_slice','_reindex_with_indexers','head',
                                '_take_with_is_copy','fillna','to_records','drop']:
                        for col in [k.name for k in result_sc.pk.fields]:
                            kensu.add_dependencies_mapping(result_sc.to_guid(),col,orig_sc.to_guid(),col,name)
                    if name in ['pivot_table']:
                        index = kwargs['index'] if "index" in kwargs else None
                        columns = kwargs['columns'] if "columns" in kwargs else None
                        values = kwargs['values'] if 'values' in kwargs else None
                        if isinstance(index, list) == False and index is not None:
                            index = list(index)
                        if isinstance(columns, list) == False and columns is not None:
                            columns = [columns]
                        if isinstance(values, list) == False and values is not None:
                            values = [values]

                        for col in result:
                            if values is not None:
                                for val in values:
                                    if len(values) == 1:
                                        kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), orig_sc.to_guid(),
                                                                       str(val), name)
                                    if val in str(col):
                                        kensu.add_dependencies_mapping(result_sc.to_guid(),str(col),orig_sc.to_guid(),val,name)

                            elif columns is not None:
                                if index is None:
                                    linked_columns = list(result.index)
                                    for element in linked_columns:
                                        kensu.add_dependencies_mapping(result_sc.to_guid(),str(col),orig_sc.to_guid(),str(element),name)
                                else:
                                    kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), orig_sc.to_guid(),str(col[0]), name)

                            elif index is not None:
                                kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), orig_sc.to_guid(),
                                                             str(col), name)
                    if name in ['_set_item']:
                        col_dest = [k.name for k in result_sc.pk.fields]
                        col_orig = [k.name for k in orig_sc.pk.fields]

                        try:
                            item = args[1]
                            orig_series = eventually_report_in_mem(kensu.extractors.extract_data_source(item, kensu.default_physical_location_ref))
                            orig_sc_series = eventually_report_in_mem(kensu.extractors.extract_schema(orig_series, item))
                        except:
                            orig_sc_series = None

                        col_series = [k.name for k in orig_sc_series.pk.fields] if orig_sc_series is not None else None
                        col_orig = list(set(col_orig) - set([args[0]]))
                        for col in col_dest:
                            if col in col_orig:
                                kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), orig_sc.to_guid(),
                                                             str(col), name)
                            elif col_series:
                                    col_origin = col_series[0]
                                    kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), orig_sc_series.to_guid(),
                                                                 str(col_origin), name)



                    if name in ['rename','_set_axis']:
                        col_dest = [k.name for k in result_sc.pk.fields]
                        col_orig = [k.name for k in orig_sc.pk.fields]
                        for i in range(len(col_dest)):
                            kensu.add_dependencies_mapping(result_sc.to_guid(), str(col_dest[i]), orig_sc.to_guid(),
                                                         str(col_orig[i]), name)

                    if name in ['merge']:
                        left = kensu_df
                        right = kwargs['right'] if 'right' in kwargs else args[0]
                        how = kwargs['how'] if 'how' in kwargs else 'inner'
                        on = kwargs['on'] if 'on' in kwargs else None
                        left_on = kwargs['left_on'] if 'left_on' in kwargs else None
                        right_on = kwargs['right_on'] if 'right_on' in kwargs else None
                        suffixes = kwargs['suffix'] if 'suffix' in kwargs else ("_x", "_y")

                        left_df = left.get_df()
                        right_df = right.get_df()

                        left_ds = orig_ds
                        left_sc = orig_sc

                        right_ds = eventually_report_in_mem(
                            kensu.extractors.extract_data_source(right_df, kensu.default_physical_location_ref,
                                                               logical_naming=kensu.logical_naming))
                        right_sc = eventually_report_in_mem(kensu.extractors.extract_schema(right_ds, right_df))


                        if how in ['inner','outer']:
                            result_cols = result.columns
                            columns_left = left_df.columns
                            columns_right = right_df.columns
                            common_columns = [value for value in columns_left if value in columns_right]
                            columns_join = on if isinstance(on, list) else [on]
                            columns_right_join = right_on if isinstance(right_on, list) else [right_on]
                            columns_left_join = left_on if isinstance(left_on, list) else [left_on]
                            suffix_left = suffixes[0]
                            suffix_right = suffixes[1]

                            combined_on = dict(zip(columns_right_join, columns_left_join))

                            for col in result_cols:
                                if col in columns_join:
                                    kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), left_sc.to_guid(),
                                                                 str(col), "Inner Join")
                                    kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), right_sc.to_guid(),
                                                                 str(col),
                                                                 "Inner Join")

                                elif col in combined_on:
                                    kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), left_sc.to_guid(),
                                                                 str(col),
                                                                 "Inner Join")
                                    kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), right_sc.to_guid(),
                                                                 combined_on[col],
                                                                 "Inner Join")

                                elif col in columns_right and col not in columns_left:
                                    kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), right_sc.to_guid(),
                                                                 str(col),
                                                                 "Inner Join")

                                elif col in columns_left and col not in columns_right:
                                    kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), left_sc.to_guid(),
                                                                 str(col),
                                                                 "Inner Join")

                                elif col.rstrip(suffix_left) in columns_left:
                                    kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), left_sc.to_guid(),
                                                                 col.rstrip(suffix_left),
                                                                 "Inner Join")

                                elif col.rstrip(suffix_right) in columns_right:
                                    kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), right_sc.to_guid(),
                                                                 col.rstrip(suffix_right), "Inner Join")

                else:
                    kensu.add_dependency((pd_df, orig_ds, orig_sc), (result, result_ds, result_sc),
                                   mapping_strategy=mapping_strategies.OUT_STARTS_WITH_IN)

            if original_result is None:
                return None
            else: 
                return result

        wrapper.__doc__ = docstring
        return wrapper

    @staticmethod
    def handle_callable_property(kensu_df, name, df, prop_value):
        logging.debug("Kensu wrapping property name=" + name)
        pd_df = kensu_df.get_df()


        if (isinstance(prop_value, _iLocIndexer) or isinstance(prop_value, _LocIndexer)) \
                and isinstance(prop_value.obj, pd.DataFrame) \
                and not isinstance(prop_value.obj, DataFrame):
            logging.debug("Wrapping dataframe in KensuDF for property name=" + name)

            dao = DataFrame.using(prop_value.obj)

            # adding the datasources, schemas and lineage to the dependencies
            if dao is not None and isinstance(dao,DataFrame):
                kensu = KensuProvider().instance()
                orig_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(pd_df, kensu.default_physical_location_ref))
                orig_sc = eventually_report_in_mem(kensu.extractors.extract_schema(orig_ds,pd_df))

                result_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(dao, kensu.default_physical_location_ref))
                result_sc = eventually_report_in_mem(kensu.extractors.extract_schema(result_ds, dao))

                if orig_ds.to_guid() != result_ds.to_guid() and orig_sc.to_guid() != result_sc.to_guid():
                    kensu.add_dependency((pd_df, orig_ds, orig_sc), (dao, result_ds, result_sc),
                                       mapping_strategy=mapping_strategies.OUT_STARTS_WITH_IN)

            # assigning the Kensu DF to the indexer to continue the tracing, wrapping the internal iloc/loc object
            prop_value.obj = dao
        else:
            # TODO log?
            pass

        return prop_value

class DataFrame(KensuPandasDelegator, pd.DataFrame):
    INTERCEPTORS = {
        re.compile('to_(.*)'): 'k_to_format',
    }


    support = KensuPandasSupport()

    __k_df = None

    def __init__(
            self,
            data=None,
            index: Optional[Axes] = None,
            columns: Optional[Axes] = None,
            dtype: Optional[Dtype] = None,
            copy: bool = False,
    ):

        if isinstance(data,ndarray):
            kensu = KensuProvider().instance()
            orig_data = data
            data = data.get_nd()
            orig_ds = eventually_report_in_mem(
                kensu.extractors.extract_data_source(orig_data, kensu.default_physical_location_ref,
                                                     logical_naming=kensu.logical_naming))
            orig_sc = eventually_report_in_mem(kensu.extractors.extract_schema(orig_ds, orig_data))

            df = pd.DataFrame(data,index,columns,dtype,copy)

            result_ds = eventually_report_in_mem(
                kensu.extractors.extract_data_source(df, kensu.default_physical_location_ref))
            result_sc = eventually_report_in_mem(kensu.extractors.extract_schema(result_ds, df))

            for col in [k.name for k in result_sc.pk.fields]:
                kensu.add_dependencies_mapping(result_sc.to_guid(), col, orig_sc.to_guid(), col, 'init')

        elif isinstance(data,dict):
            from kensu.itertools import kensu_list
            kensu = KensuProvider().instance()
            deps ={}
            for key in data.keys():
                item = data[key]

                if isinstance(item,kensu_list):
                    deps[key] = item.deps

                if isinstance(item,ndarray):
                    orig_ds = eventually_report_in_mem(
                        kensu.extractors.extract_data_source(item, kensu.default_physical_location_ref))
                    orig_sc = eventually_report_in_mem(kensu.extractors.extract_schema(orig_ds, item))
                    deps[key] = [orig_sc]


            df = pd.DataFrame(data, index, columns, dtype, copy)
            result_ds = eventually_report_in_mem(
                kensu.extractors.extract_data_source(df, kensu.default_physical_location_ref))
            result_sc = eventually_report_in_mem(kensu.extractors.extract_schema(result_ds, df))

            for key in deps.keys():
                for orig_sc in deps[key]:
                    for col in [k.name for k in orig_sc.pk.fields]:
                        kensu.add_dependencies_mapping(result_sc.to_guid(), str(key), orig_sc.to_guid(), str(col), 'init')



        else:
            df = pd.DataFrame(data, index, columns, dtype, copy)

        self.kensu_init(df)


    @staticmethod
    def using(df):
        if isinstance(df, DataFrame):
            d = DataFrame.using(df.get_df())
            return d
        else:
            d = DataFrame()
            d.kensu_init(df)
            return d

    def get_df(self):
        # if the current Kensu DataFrame isn't delegating... TODO.. need to have a type for this instead...
        if self.__k_df is None:
            return self
        else:
            return self.__k_df

    def kensu_init(self, df):
        self.__k_df = df


    buffer_param_name_re = re.compile('.*buf(?:fer)?')

    def k_to_format(self, result_regex, *args, **kwargs):
        kensu = KensuProvider().instance()
        orig_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(self.get_df(), kensu.default_physical_location_ref,logical_naming=kensu.logical_naming))
        orig_sc = eventually_report_in_mem(kensu.extractors.extract_schema(orig_ds, self.get_df()))

        # the regex for this method has a group to get the read format
        fmt = result_regex.group(1)

        # look up for buf
        location = None
        for param_name in kwargs.keys():
            if self.buffer_param_name_re.search(param_name) is not None:
                location = kwargs[param_name]

        #FIXME
        table = None
        if fmt == 'gbq' or isinstance(args[0],str) == False:
            return None
        if fmt == 'sql':
            from sqlalchemy.engine.base import Engine
            engine = None
            for arg in args:
                if isinstance(arg,Engine):
                    engine = arg
            if engine == None:
                engine = kwargs['con']
            if engine is not None and engine.name == 'postgresql':
                fmt = 'Postgres table'

                #FIXME How to find the schema name?
                location = "postgres://"+ engine.url.host +":"+ str(engine.url.port)+'/'+engine.url.database + '.public.' + args[0]

        if location is None and location != 'BigQuery Table' and len(args) > 0:
            location = get_absolute_path(args[0])

        if location is not None and fmt is not None:
            df = self.get_df()

            #FIXME generify for written target ds (BigQuery, mysql, postgre...), same for location and format
            ds = kensu.extractors.extract_data_source(self.get_df(), kensu.default_physical_location_ref, location=location, format=fmt,logical_naming=kensu.logical_naming)._report()
            if fmt == 'BigQuery Table':
                sc = kensu.extractors.extract_schema(ds, table)._report()
            else:
                sc = kensu.extractors.extract_schema(ds, self.get_df())._report()

            kensu.real_schema_df[sc.to_guid()] = df

            # add dep between original data source and written one
            if kensu.mapping :
                for col in df:
                    kensu.add_dependencies_mapping(sc.to_guid(),str(col),orig_sc.to_guid(),str(col),'Write')
            else:
                kensu.add_dependency((df, orig_ds, orig_sc), (df, ds, sc), mapping_strategy=mapping_strategies.DIRECT)

            if kensu.mapping:
                kensu.report_with_mapping()

            else:
                def process_dag(out):

                    deps = kensu.get_dependencies()

                    def process_dag_layer(out, layer="-"):
                        for [i, o, strategy] in iter(deps):
                            in_ds_id = KensuClassHandlers.guid(i[1])
                            logging.debug(layer +"dep.in.datasource.id {}".format(str(in_ds_id)))
                            out_ds_id = KensuClassHandlers.guid(o[1])
                            logging.debug(layer +"dep.out.datasource.id {}".format(str(out_ds_id)))
                            current_ds_id = KensuClassHandlers.guid(out[1])
                            logging.debug(layer +"current.datasource.id {}".format(str(current_ds_id)))
                            in_sc_id = KensuClassHandlers.guid(i[2])
                            logging.debug(layer +"dep.in.schema.id {}".format(str(in_sc_id)))
                            out_sc_id = KensuClassHandlers.guid(o[2])
                            logging.debug(layer +"dep.out.schema.id {}".format(str(out_sc_id)))
                            current_sc_id = KensuClassHandlers.guid(out[2])
                            logging.debug(layer +"current.schema.id {}".format(str(current_sc_id)))
                            logging.debug("-------------------")
                            if out_ds_id == current_ds_id and out_sc_id == current_sc_id:
                                if isinstance(i[0], DataFrame):
                                    i = (i[0].get_df(), i[1], i[2])
                                if isinstance(o[0], DataFrame):
                                    o = (o[0].get_df(), o[1], o[2])
                                lin = kensu.s.n.with_input(i) \
                                    .with_output(o) \
                                    .with_strategy(strategy) \
                                    .e
                                lin.e()

                                # try process previous layer
                                # FIXME.. need to handle loops => I guess this needs to be solved in the dependencies itself
                                process_dag_layer(i, "  " + layer)
                        return
                    process_dag_layer(out)
                    return
                process_dag((self.get_df(), ds, sc))

class KensuSeriesDelegator(object):
    SKIP_KENSU_FIELDS = ["_Series__k_s", "INTERCEPTORS"]
    SKIP_KENSU_METHODS = ["get_s", "kensu_init", "k_to_format", "to_string"]

    def __getattribute__(self, name):
        if name in ['str',"to_string"]:
            attr_value = object.__getattribute__(self.get_s(), name)
        else:
            attr_value = object.__getattribute__(self, name)

        if name == "_Series__k_s":
            return attr_value
        if name == "__class__":
            return attr_value
        if hasattr(attr_value, '__call__'):
            return KensuSeriesDelegator.handle_callable(self, name, attr_value)
        else:
            return KensuSeriesDelegator.handle_field(self, name, attr_value)

    def wrapped_series_op(self, other_input, wrapped_fn, op_title=None):
        kensu = KensuProvider().instance()
        op_title = op_title or 'Series ' + str(wrapped_fn.__name__)
        input_kensu_series = self

        other_series = other_input.get_s() if isinstance(other_input,Series) else other_input
        result = Series.using(wrapped_fn(other_series))
        result_ds = eventually_report_in_mem(
            kensu.extractors.extract_data_source(result, kensu.default_physical_location_ref,
                                                 logical_naming=kensu.logical_naming))
        result_sc = eventually_report_in_mem(kensu.extractors.extract_schema(result_ds, result))

        inputs = [self,other_input]
        for input in inputs:
            if isinstance(input,Series):
                input_ds = eventually_report_in_mem(
                    kensu.extractors.extract_data_source(input, kensu.default_physical_location_ref,
                                                         logical_naming=kensu.logical_naming))
                input_sc = eventually_report_in_mem(kensu.extractors.extract_schema(input_ds, input))
                result_col = [k.name for k in result_sc.pk.fields][0]
                orig_col = [k.name for k in input_sc.pk.fields][0]
                kensu.add_dependencies_mapping(result_sc.to_guid(), str(result_col), input_sc.to_guid(),
                                               str(orig_col), str(wrapped_fn.__name__))
        return result

    def __sub__(self, other):
         return self.wrapped_series_op(other, self.get_s().__sub__)

    @staticmethod
    def wrap_returned_df(kensu_s, returned, name, original):
        if isinstance(returned, pd.Series) \
                and not isinstance(returned, Series) \
                and original is not returned:
            logging.debug("Wrapping attr access resulting pd.Series in KensuS for property name=" + name)
            tmp = Series.using(returned)
            returned = tmp

        elif isinstance(returned, numpy.ndarray):
            logging.debug("Wrapping attr access resulting numpy.ndarray in KensuS for property name=" + name)
            tmp = ndarray.using(returned)
            returned = tmp

        return returned

    @staticmethod
    def handle_field(kensu_s, name, attr_value):
        # _Series__k_s => points to Series.__k_s
        if name in KensuSeriesDelegator.SKIP_KENSU_FIELDS:
            # return Kensu DF attr value right away
            logging.debug("Kensu attribute name=" + name)
            return attr_value
        else:
            logging.debug("NOT Kensu, let's delegate attribute name=" + name)

            pd_s = kensu_s.get_s()
            s_attr = object.__getattribute__(pd_s, name)
            result = KensuSeriesDelegator.wrap_returned_df(kensu_s, s_attr, name, pd_s)


            #adding the datasources, schemas and lineage to the dependencies
            if result is not None and isinstance(result, Series):
                kensu = KensuProvider().instance()
                orig_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(pd_s, kensu.default_physical_location_ref))
                orig_sc = eventually_report_in_mem(kensu.extractors.extract_schema(orig_ds,pd_s))

                result_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(result, kensu.default_physical_location_ref))
                result_sc = eventually_report_in_mem(kensu.extractors.extract_schema(result_ds, result))

                kensu.add_dependency((pd_s, orig_ds, orig_sc), (result, result_ds, result_sc),
                                   mapping_strategy=mapping_strategies.OUT_STARTS_WITH_IN)

            elif result is not None and isinstance(result, ndarray):
                kensu = KensuProvider().instance()
                orig_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(pd_s, kensu.default_physical_location_ref))
                orig_sc = eventually_report_in_mem(kensu.extractors.extract_schema(orig_ds,pd_s))

                result_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(result, kensu.default_physical_location_ref))
                result_sc = eventually_report_in_mem(kensu.extractors.extract_schema(result_ds, result))

                if kensu.mapping:
                    result_col = [k.name for k in result_sc.pk.fields][0]
                    orig_col = [k.name for k in orig_sc.pk.fields][0]
                    kensu.add_dependencies_mapping(result_sc.to_guid(), str(result_col), orig_sc.to_guid(),
                                                   str(orig_col), name)

                else:
                    kensu.add_dependency((pd_s, orig_ds, orig_sc), (result, result_ds, result_sc),
                                   mapping_strategy=mapping_strategies.OUT_STARTS_WITH_IN)

            elif result is not None and isinstance(result, StringMethods):

                class Kensu_StringMethods(StringMethods):

                    def __getitem__(self, key):
                        if isinstance(key, slice):
                            original_result = self.slice(start=key.start, stop=key.stop, step=key.step)
                        else:
                            original_result = self.get(key)

                        orig_series = self._parent

                        kensu = KensuProvider().instance()

                        orig_ds = kensu.extractors.extract_data_source(orig_series, kensu.default_physical_location_ref)
                        orig_sc = kensu.extractors.extract_schema(orig_ds, orig_series)
                        res_ds = kensu.extractors.extract_data_source(original_result,
                                                                      kensu.default_physical_location_ref)
                        res_sc = kensu.extractors.extract_schema(res_ds, original_result)
                        result_cols = [k.name for k in res_sc.pk.fields]
                        for col in result_cols:
                            kensu.add_dependencies_mapping(res_sc.to_guid(), str(col), orig_sc.to_guid(), str(col), '__getitem__')

                        return Series.using(original_result)
                result.__class__ = Kensu_StringMethods

            elif result is not None and isinstance(result,CategoricalAccessor):
                class Kensu_CategoricalAccessor(CategoricalAccessor):
                    origin = None
                    def set_origin(self,orig):
                        self.origin = orig

                    def get_origin(self):
                        return self.origin

                    @property
                    def codes(self):
                        """
                        Return Series of codes as well as the index.
                        """
                        kensu = KensuProvider().instance()
                        result = pd.Series(self._parent.codes, index=self._index)

                        origin = self.get_origin()

                        orig_ds = kensu.extractors.extract_data_source(origin, kensu.default_physical_location_ref)
                        orig_sc = kensu.extractors.extract_schema(orig_ds, origin)

                        res_ds = kensu.extractors.extract_data_source(result,
                                                                      kensu.default_physical_location_ref)
                        res_sc = kensu.extractors.extract_schema(res_ds, result)
                        result_col = [k.name for k in res_sc.pk.fields][0]
                        orig_col = [k.name for k in orig_sc.pk.fields][0]
                        kensu.add_dependencies_mapping(res_sc.to_guid(), str(result_col), orig_sc.to_guid(),
                                                       str(orig_col), 'code')

                        return Series.using(result)


                result.__class__ = Kensu_CategoricalAccessor
                result.set_origin(kensu_s)

            return result

    @staticmethod
    def handle_callable(kensu_s, name, attr_value):
        if name not in KensuSeriesDelegator.SKIP_KENSU_METHODS:
            pd_s = object.__getattribute__(kensu_s, "get_s")()
            # for example, `iloc` is subscriptable and is a property, hence call this way: iloc[]
            if isinstance(getattr(type(pd_s), name), property):
                delegated_pd_attr = object.__getattribute__(pd_s, name)
                return KensuSeriesDelegator.handle_callable_property(kensu_s, name, pd_s, delegated_pd_attr)
            else:
                delegated_pd_attr = object.__getattribute__(pd_s, name)
                docstring = delegated_pd_attr.__doc__
                return KensuSeriesDelegator.create_function_wrapper(kensu_s, name, docstring)
        else:
            logging.debug("Kensu function name=" + name)

            # return Kensu DF function result right away
            def w(*args, **kwargs):
                return attr_value(*args, **kwargs)
            return w

    @staticmethod
    def create_function_wrapper(kensu_s, name, docstring):
        logging.debug("Kensu wrapping function name=" + name)

        def wrapper(*args, **kwargs):
            # call method on delegated df
            delegation_config = getattr(kensu_s, 'INTERCEPTORS', None)

            for regex_attr_name in delegation_config:
                result_regex = regex_attr_name.search(name)
                if result_regex is not None:
                    # if we pass the regex, there is an interception defined, so we execute it
                    name_of_intercept_method = delegation_config[regex_attr_name]
                    intercept = object.__getattribute__(kensu_s, name_of_intercept_method)
                    intercept(result_regex, *args, **kwargs)

            pd_s = kensu_s.get_s()

            inplace = False
            if 'inplace' in kwargs:
                if kwargs['inplace']:
                    inplace = True
                    cacher = getattr(pd_s, "_cacher", None)
                    if cacher is not None:
                        kensu = KensuProvider().instance()
                        origin_ds = cacher[1]()
                        orig_ds = kensu.extractors.extract_data_source(origin_ds, kensu.default_physical_location_ref)
                        orig_sc = kensu.extractors.extract_schema(orig_ds, origin_ds)
                        origin_cols = [k.name for k in orig_sc.pk.fields]



            s_attr = object.__getattribute__(pd_s, name)

            result = s_attr(*args, **kwargs)
            original_result = result

            result = KensuSeriesDelegator.wrap_returned_df(kensu_s, result, name, pd_s)

            if result is None :
                #set_axis updates the Series directly (self) and returns None, so we reset the result
                if name in ["fillna"]:
                    result = kensu_s

            if inplace:
                cacher = getattr(pd_s, "_cacher", None)
                if cacher is not None:
                    result_ds = cacher[1]()
                    res_ds = kensu.extractors.extract_data_source(result_ds, kensu.default_physical_location_ref)
                    res_sc = kensu.extractors.extract_schema(res_ds, result_ds)
                    result_cols = [k.name for k in res_sc.pk.fields]

                    for col in result_cols:
                        kensu.add_dependencies_mapping(res_sc.to_guid(), str(col), orig_sc.to_guid(), str(col), name)

                    return original_result

            #adding the datasources, schemas and lineage to the dependencies
            if result_regex is None and result is not None and (isinstance(result,DataFrame) or isinstance(result,Series)):
                kensu = KensuProvider().instance()
                orig_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(pd_s, kensu.default_physical_location_ref))
                orig_sc = eventually_report_in_mem(kensu.extractors.extract_schema(orig_ds,pd_s))

                result_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(result, kensu.default_physical_location_ref))
                result_sc = eventually_report_in_mem(kensu.extractors.extract_schema(result_ds, result))

                if kensu.mapping:
                    if name in ['fillna','replace','astype','_get_with']:
                        for col in [k.name for k in result_sc.pk.fields]:
                            kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), orig_sc.to_guid(), str(col), name)

                else:
                    kensu.add_dependency((pd_s, orig_ds, orig_sc), (result, result_ds, result_sc),
                                       mapping_strategy=mapping_strategies.OUT_STARTS_WITH_IN)

            if original_result is None:
                return None
            else:
                return result

        wrapper.__doc__ = docstring
        return wrapper

    @staticmethod
    def handle_callable_property(kensu_s, name, s, prop_value):
        logging.debug("Kensu wrapping property name=" + name)
        pd_s = kensu_s.get_s()

        # FIXME what shall we do here? => There is some lineage to build. Example: iloc
        # FIXME: _LocationIndexer only? limitations?

        if (isinstance(prop_value, _iLocIndexer) or isinstance(prop_value, _LocIndexer)) \
                and isinstance(prop_value.obj, pd.DataFrame) \
                and not isinstance(prop_value.obj, DataFrame):
            logging.debug("Wrapping dataframe in KensuDF for property name=" + name)
            dao = DataFrame.using(prop_value.obj)
            # adding the datasources, schemas and lineage to the dependencies
            if dao is not None and isinstance(dao,DataFrame):
                kensu = KensuProvider().instance()
                orig_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(pd_s, kensu.default_physical_location_ref))
                orig_sc = eventually_report_in_mem(kensu.extractors.extract_schema(orig_ds, pd_s))

                result_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(dao, kensu.default_physical_location_ref))
                result_sc = eventually_report_in_mem(kensu.extractors.extract_schema(result_ds, dao))

                if orig_ds.to_guid() != result_ds.to_guid() and orig_sc.to_guid() != result_sc.to_guid():
                    kensu.add_dependency((pd_s, orig_ds, orig_sc), (dao, result_ds, result_sc),
                                       mapping_strategy=mapping_strategies.OUT_STARTS_WITH_IN)

            # assigning the Kensu DF to the indexer to continue the tracing
            prop_value.obj = dao
        else:
            # TODO log?
            pass

        return prop_value

class Series(KensuSeriesDelegator, pd.Series):
    INTERCEPTORS = {
        re.compile('to_(.*)'): 'k_to_format',
    }

    # TODO support = KensuPandasSupport()

    __k_s = None


    @staticmethod
    def using(s):
        if isinstance(s, Series):
            cs = Series.using(s.get_s())
            return cs
        else:
            cs = Series()
            cs.kensu_init(s)
            return cs

    def get_s(self):
        # if the current Kensu Series isn't delegating... TODO.. need to have a type for this instead...
        if self.__k_s is None:
            return self
        else:
            return self.__k_s

    def kensu_init(self, s):
        self.__k_s = s

    buffer_param_name_re = re.compile('.*buf(?:fer)?')

    def __add__(self, other):
        #FIXME Something similar to ops/__init.py -> _arith_method_SERIES should be done t avoid code duplication
        left = self.get_s()
        right = other.get_s()
        result = left+right

        '''
        kensu = KensuProvider().instance()

        orig_ds = eventually_report_in_mem(
            kensu.extractors.extract_data_source(self.get_s(), kensu.default_physical_location_ref,
                                               logical_naming=kensu.logical_naming))

        orig_sc = eventually_report_in_mem(kensu.extractors.extract_schema(orig_ds, self.get_s()))
        '''

        #TODO Link

        if isinstance(result, Series):
            result = Series.using(result)
        elif isinstance(result,ndarray):
            result = ndarray.using(result)
        else:
            result = result
            #TODO log
        return result

    def k_to_format(self, result_regex, *args, **kwargs):
        kensu = KensuProvider().instance()

        orig_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(self.get_s(), kensu.default_physical_location_ref,logical_naming=kensu.logical_naming))

        orig_sc = eventually_report_in_mem(kensu.extractors.extract_schema(orig_ds, self.get_s()))

        # the regex for this method has a group to get the read format
        fmt = result_regex.group(1)

        # look up for buf
        location = None
        for param_name in kwargs.keys():
            if self.buffer_param_name_re.search(param_name) is not None:
                location = kwargs[param_name]

        if location is None and len(args) > 0:

            location = get_absolute_path(args[0])

        if location is not None and fmt is not None:
            s = self.get_s()

            ds = kensu.extractors.extract_data_source(self.get_s(), kensu.default_physical_location_ref, location=location, format=fmt,logical_naming=kensu.logical_naming)._report()

            sc = kensu.extractors.extract_schema(ds, self.get_s())._report()

            if kensu.mapping :
                kensu.add_dependencies_mapping(sc.to_guid(),self.get_s().name,orig_sc.to_guid(),self.get_s().name,'Write')
                kensu.real_schema_df[sc.to_guid()]=self.get_s()
                kensu.report_with_mapping()
            else:
                # add dep between original data source and written one
                kensu.add_dependency((s, orig_ds, orig_sc), (s, ds, sc), mapping_strategy=mapping_strategies.DIRECT)

                def process_dag(out):
                    deps = kensu.get_dependencies()

                    def process_dag_layer(out, layer="-"):
                        for [i, o, strategy] in iter(deps):
                            in_ds_id = KensuClassHandlers.guid(i[1])
                            logging.debug(layer +"dep.in.datasource.id {}".format(str(in_ds_id)))
                            out_ds_id = KensuClassHandlers.guid(o[1])
                            logging.debug(layer +"dep.out.datasource.id {}".format(str(out_ds_id)))
                            current_ds_id = KensuClassHandlers.guid(out[1])
                            logging.debug(layer +"current.datasource.id {}".format(str(current_ds_id)))
                            in_sc_id = KensuClassHandlers.guid(i[2])
                            logging.debug(layer +"dep.in.schema.id {}".format(str(in_sc_id)))
                            out_sc_id = KensuClassHandlers.guid(o[2])
                            logging.debug(layer +"dep.out.schema.id {}".format(str(out_sc_id)))
                            current_sc_id = KensuClassHandlers.guid(out[2])
                            logging.debug(layer +"current.schema.id {}".format(str(current_sc_id)))
                            logging.debug("-------------------")
                            if out_ds_id == current_ds_id and out_sc_id == current_sc_id:
                                if isinstance(i[0], DataFrame):
                                    i = (i[0].get_df(), i[1], i[2])
                                elif isinstance(i[0], Series):
                                    i = (i[0].get_s(), i[1], i[2])
                                if isinstance(o[0], DataFrame):
                                    o = (o[0].get_df(), o[1], o[2])
                                elif isinstance(o[0], Series):
                                    o = (o[0].get_s(), o[1], o[2])
                                lin = kensu.s.n.with_input(i) \
                                    .with_output(o) \
                                    .with_strategy(strategy) \
                                    .e
                                lin.e()

                                # try process previous layer
                                # FIXME.. need to handle loops => I guess this needs to be solved in the dependencies itself
                                process_dag_layer(i, "  " + layer)
                        return
                    process_dag_layer(out)
                return
                process_dag((self.get_s(), ds, sc))

def wrap_pandas_reader(reader):
    def wrapper(*args, **kwargs):
        kensu = KensuProvider().instance()

        if reader.__name__.startswith("read_"):
            fmt = reader.__name__[len("read_"):]
        else:
            fmt = "unknown"

        df = reader(*args, **kwargs)
        read_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(df, kensu.default_physical_location_ref,logical_naming=kensu.logical_naming))
        read_sc = eventually_report_in_mem(kensu.extractors.extract_schema(read_ds, df))


        # TODO... I think other param names are expected here for other reader functions
        # prepare original data source
        #TODO: handle url,s3,etc
        if "filepath_or_buffer" in kwargs:
            location = get_absolute_path(kwargs["filepath_or_buffer"])
        elif args[0].startswith('http://') or args[0].startswith('https://'):
            location = args[0]
        else:
            location = get_absolute_path(args[0])

        df_kensu = DataFrame.using(df)

        ds = kensu.extractors.extract_data_source(df_kensu, kensu.default_physical_location_ref, location=location, format=fmt,logical_naming=kensu.logical_naming)._report()

        sc = kensu.extractors.extract_schema(ds, df_kensu)._report()
        kensu.real_schema_df[sc.to_guid()] = df
        if kensu.mapping == True:
            for col in df:
                kensu.add_dependencies_mapping(read_sc.to_guid(), str(col), sc.to_guid(), str(col),
                                             "Read")
        else:
            kensu.add_dependency((df, ds, sc), (df, read_ds, read_sc),mapping_strategy = mapping_strategies.OUT_STARTS_WITH_IN)

        return df_kensu

    wrapper.__doc__ = reader.__doc__
    return wrapper

def wrap_pandas_get_dummies(method):
    def wrapper(*args, **kwargs):
        kensu = KensuProvider().instance()
        df_result = method(*args, **kwargs)
        df = args[0] # see get_dummies definition (first arg is `data`)

        orig_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(df, kensu.default_physical_location_ref,logical_naming=kensu.logical_naming))
        orig_sc = eventually_report_in_mem(kensu.extractors.extract_schema(orig_ds, df))

        result_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(df_result, kensu.default_physical_location_ref,logical_naming=kensu.logical_naming))
        result_sc = eventually_report_in_mem(kensu.extractors.extract_schema(result_ds, df_result))



        df_result_kensu = DataFrame.using(df_result)

        if kensu.mapping:
            col_dest = [k.name for k in result_sc.pk.fields]
            col_orig = [k.name for k in orig_sc.pk.fields]
            prefix = kwargs['prefix'] if 'prefix' in kwargs else None
            prefix_sep = kwargs['prefix_sep'] if 'prefix_sep' in kwargs else "_"
            columns = kwargs['columns'] if 'columns' in kwargs else None

            for col in col_dest:
                if col in col_orig:
                    kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), orig_sc.to_guid(), str(col),
                                                 "Dummy")
                else:
                    origin_col = col.split(prefix_sep)[0]
                    if prefix:
                        if isinstance(prefix,list):
                            index = prefix.index(origin_col)
                            origin_col = columns[index]

                    kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), orig_sc.to_guid(), str(origin_col),
                                                 "Dummy")

        else:
            kensu.add_dependency((df,orig_ds,orig_sc),(df_result,result_ds,result_sc),mapping_strategy=mapping_strategies.OUT_STARTS_WITH_IN)

        return df_result_kensu

    wrapper.__doc__ = method.__doc__
    return wrapper

def wrap_merge(method):
    def wrapper(*args, **kwargs):
        kensu = KensuProvider().instance()

        new_args = []
        for item in args:
            if isinstance(item, DataFrame):
                new_args.append(item.get_df())
            elif isinstance(item, Series):
                new_args.append(item.get_s())
            else:
                new_args.append(item)
        new_args = tuple(new_args)

        new_kwargs = {}
        for item in kwargs:
            if isinstance(kwargs[item], DataFrame):
                new_kwargs[item]=kwargs[item].get_df()
            elif isinstance(item, Series):
                new_kwargs[item]=kwargs[item].get_s()
            else:
                new_kwargs[item]=kwargs[item]

        df_result = method(*new_args, **new_kwargs)

        left = kwargs['left'] if 'left' in kwargs else args[0]
        right = kwargs['right'] if 'right' in kwargs else args[1]
        how =  kwargs['how'] if 'how' in kwargs else 'inner'
        on = kwargs['on'] if 'on' in kwargs else None
        left_on = kwargs['left_on'] if 'left_on' in kwargs else None
        right_on = kwargs['right_on'] if 'right_on' in kwargs else None
        suffixes = kwargs['suffix'] if 'suffix' in kwargs else ("_x", "_y")

        left_df = left.get_df()
        right_df = right.get_df()

        left_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(left_df, kensu.default_physical_location_ref,
                                                                              logical_naming=kensu.logical_naming))
        left_sc = eventually_report_in_mem(kensu.extractors.extract_schema(left_ds, left_df))

        right_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(right_df, kensu.default_physical_location_ref,
                                                                              logical_naming=kensu.logical_naming))
        right_sc = eventually_report_in_mem(kensu.extractors.extract_schema(right_ds, right_df))


        result_ds = eventually_report_in_mem(
            kensu.extractors.extract_data_source(df_result, kensu.default_physical_location_ref,
                                               logical_naming=kensu.logical_naming))
        result_sc = eventually_report_in_mem(kensu.extractors.extract_schema(result_ds, df_result))

        df_result_kensu = DataFrame.using(df_result)

        if kensu.mapping:
            if how == 'inner':
                result_cols = df_result.columns
                columns_left = left_df.columns
                columns_right = right_df.columns
                common_columns = [value for value in columns_left if value in columns_right]
                columns_join = on if isinstance(on,list) else [on]
                columns_right_join = right_on if isinstance(right_on,list) else [right_on]
                columns_left_join = left_on if isinstance(left_on, list) else [left_on]
                suffix_left = suffixes[0]
                suffix_right = suffixes[1]

                combined_on = dict(zip(columns_right_join,columns_left_join))

                for col in result_cols:
                    if col in columns_join:
                        kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), left_sc.to_guid(), str(col), "Inner Join")
                        kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), right_sc.to_guid(), str(col),
                                                     "Inner Join")

                    elif col in combined_on:
                        kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), left_sc.to_guid(), str(col),
                                                     "Inner Join")
                        kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), right_sc.to_guid(), combined_on[col],
                                                     "Inner Join")

                    elif col in columns_right and col not in columns_left:
                        kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), right_sc.to_guid(), str(col),
                                                     "Inner Join")

                    elif col in columns_left and col not in columns_right:
                        kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), left_sc.to_guid(), str(col),
                                                     "Inner Join")

                    elif col.rstrip(suffix_left) in columns_left:
                        kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), left_sc.to_guid(), col.rstrip(suffix_left),
                                                    "Inner Join")

                    elif col.rstrip(suffix_right) in columns_right:
                        kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), right_sc.to_guid(),
                                                     col.rstrip(suffix_right),"Inner Join")

        return df_result_kensu

    wrapper.__doc__ = method.__doc__
    return wrapper


def wrap_external_to_pandas_transformation(method, get_inputs_lineage_fn):
    def wrapper(*args, **kwargs):
        kensu = KensuProvider().instance()

        ext_inputs_lineage = get_inputs_lineage_fn(kensu, *args, **kwargs)  # type: GenericComputedInMemDs
        df_result = method(*args, **kwargs)

        ext_inputs_lineage.report(ksu=kensu,
                                  df_result=df_result,
                                  operation_type="sparkDf.toPandas()",
                                  report_output=kensu.report_in_mem)
        result_ksu_pandas_df = DataFrame.using(df_result)
        return result_ksu_pandas_df

    wrapper.__doc__ = method.__doc__
    return wrapper

def wrap_to_datetime(method):
    def wrapper(*args, **kwargs):
        kensu = KensuProvider().instance()

        new_args = []
        for item in args:
            if isinstance(item, DataFrame):
                new_args.append(item.get_df())
            elif isinstance(item, Series):
                new_args.append(item.get_s())
            else:
                new_args.append(item)
        new_args = tuple(new_args)

        new_kwargs = {}
        for item in kwargs:
            if isinstance(kwargs[item], DataFrame):
                new_kwargs[item]=kwargs[item].get_df()
            elif isinstance(item, Series):
                new_kwargs[item]=kwargs[item].get_s()
            else:
                new_kwargs[item]=kwargs[item]

        df_result = method(*new_args, **new_kwargs)


        result_ds = eventually_report_in_mem(
            kensu.extractors.extract_data_source(df_result, kensu.default_physical_location_ref,
                                               logical_naming=kensu.logical_naming))
        result_sc = eventually_report_in_mem(kensu.extractors.extract_schema(result_ds, df_result))

        for orig_df in args:
            orig_ds = eventually_report_in_mem(
                kensu.extractors.extract_data_source(orig_df, kensu.default_physical_location_ref,
                                                     logical_naming=kensu.logical_naming))
            orig_sc = eventually_report_in_mem(kensu.extractors.extract_schema(orig_ds, orig_df))

            for col in [k.name for k in orig_sc.pk.fields]:
                kensu.add_dependencies_mapping(result_sc.to_guid(), col, orig_sc.to_guid(), col, 'to_datetime')

        df_result_kensu = Series.using(df_result)

        return df_result_kensu

    wrapper.__doc__ = method.__doc__
    return wrapper

def wrap_concat(method):
    def wrapper(*args, **kwargs):
        kensu = KensuProvider().instance()
        df_result = method(*args, **kwargs)
        result_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(df_result, kensu.default_physical_location_ref,logical_naming=kensu.logical_naming))
        result_sc = eventually_report_in_mem(kensu.extractors.extract_schema(result_ds, df_result))

        col_dest = [k.name for k in result_sc.pk.fields]

        for df in args[0]:
            orig_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(df, kensu.default_physical_location_ref,logical_naming=kensu.logical_naming))
            orig_sc = eventually_report_in_mem(kensu.extractors.extract_schema(orig_ds, df))
            col_orig = [k.name for k in orig_sc.pk.fields]
            for col in col_dest:
                if col in col_orig:
                    kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), orig_sc.to_guid(), str(col),
                                           "concat")
        df_result_kensu = DataFrame.using(df_result)
        return df_result_kensu

    wrapper.__doc__ = method.__doc__
    return wrapper



from typing import TYPE_CHECKING

import logging

from airflow.providers.google.cloud.hooks.gcs import GCSHook
from airflow.providers.google.cloud.transfers import gcs_to_gcs as airflow_gcs_to_gcs

from kensu.airflow.kensu_airflow_collector import COLLECTOR_STATUS_INIT, COLLECTOR_STATUS_DONE, log_status, \
    airflow_init_kensu, handle_ex
from kensu.utils.dsl.extractors.external_lineage_dtos import GenericComputedInMemDs

if TYPE_CHECKING:
    from airflow.utils.context import Context


class GCSToGCSOperator(airflow_gcs_to_gcs.GCSToGCSOperator):
    def __init__(
            self,
            *args,
            **kwargs,
    ):
        super(GCSToGCSOperator, self).__init__(
            *args, **kwargs
        )
        self.ksu_copied_objects = []

    def _copy_single_object(self, hook, source_object, destination_object):
        res = super(GCSToGCSOperator, self)._copy_single_object(hook, source_object, destination_object)
        logging.info(
            f"Kensu _copy_single_object(source_object={source_object}, destination_object={destination_object})")
        self.ksu_copied_objects.append([source_object, destination_object])
        return res

    def _is_dir(self, hook, bucket_name, object_name):
        try:
            hook.get_blob_update_time(bucket_name, object_name)
            return False
        except ValueError:
            return True

    def execute(self, context: 'Context') -> None:
        # FIMXE: there's `self.move_object: bool`, we might want to get some info (stats) before actual copy happens
        res = super(GCSToGCSOperator, self).execute(context)
        try:
            log_status(self, COLLECTOR_STATUS_INIT)
            airflow_init_kensu(airflow_operator=self)
            logging.info(f"self.ksu_copied_objects={self.ksu_copied_objects}")
            hook = GCSHook(
                gcp_conn_id=self.gcp_conn_id,
                delegate_to=self.delegate_to,
                impersonation_chain=self.impersonation_chain,
            )
            for src, dest in self.ksu_copied_objects:
                src_uri = f"gs://{self.source_bucket}/{src}"
                dest_uri = f"gs://{self.destination_bucket}/{dest}"
                dest_categories = None
                # FIXME: do we any need special LDSo/PDSo logic here?
                # if self.destination_object and \
                #         self._is_dir(hook, bucket_name=self.destination_bucket, object_name=self.destination_object):
                #     lds_relative_path = self.destination_object.rstrip('/')
                #     lds_name = f"gs :: {lds_relative_path}"
                #     lds_location = full_gs_uri(self.destination_bucket, lds_relative_path)
                #     lds_name = lds_location
                #     dest_categories = [
                #         f"logical::{lds_name}",
                #         f"logicalLocation::{lds_location}"
                #     ]

                # e.g.:
                #             source_object=f'{INPUT_PART}/{colour}_{DATASET}*.{INPUT_FILETYPE}',
                #             destination_object=f'{colour}/{colour}_{DATASET}',
                logging.info(f"""
                GCSToGCSOperator:
                    src={src_uri},
                    dest={dest_uri},
                    dest_categories={dest_categories},
                    """)
                GenericComputedInMemDs.report_copy_with_opt_schema(
                    src=src_uri,
                    dest=dest_uri,
                    dest_categories=dest_categories,
                    operation_type='GCSToGCSOperator'
                )
        except Exception as ex:
            handle_ex(self, ex)
        log_status(self, COLLECTOR_STATUS_DONE)
        return res


GCSToGCSOperator.__doc__ = airflow_gcs_to_gcs.GCSToGCSOperator.__doc__



# coding: utf-8

"""
    

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)

    OpenAPI spec version: beta
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


from pprint import pformat

from six import iteritems


class BatchErrorMessage(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """


    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'error_message': 'str',
        'entity_type': 'str',
        'entity_json': 'str'
    }

    attribute_map = {
        'error_message': 'errorMessage',
        'entity_type': 'entityType',
        'entity_json': 'entityJson'
    }

    def __init__(self, error_message=None, entity_type=None, entity_json=None):
        """
        BatchErrorMessage - a model defined in Swagger
        """

        self._error_message = None
        self._entity_type = None
        self._entity_json = None

        self.error_message = error_message
        self.entity_type = entity_type
        self.entity_json = entity_json

    @property
    def error_message(self):
        """
        Gets the error_message of this BatchErrorMessage.

        :return: The error_message of this BatchErrorMessage.
        :rtype: str
        """
        return self._error_message

    @error_message.setter
    def error_message(self, error_message):
        """
        Sets the error_message of this BatchErrorMessage.

        :param error_message: The error_message of this BatchErrorMessage.
        :type: str
        """
        if error_message is None:
            raise ValueError("Invalid value for `error_message`, must not be `None`")

        self._error_message = error_message

    @property
    def entity_type(self):
        """
        Gets the entity_type of this BatchErrorMessage.

        :return: The entity_type of this BatchErrorMessage.
        :rtype: str
        """
        return self._entity_type

    @entity_type.setter
    def entity_type(self, entity_type):
        """
        Sets the entity_type of this BatchErrorMessage.

        :param entity_type: The entity_type of this BatchErrorMessage.
        :type: str
        """
        if entity_type is None:
            raise ValueError("Invalid value for `entity_type`, must not be `None`")

        self._entity_type = entity_type

    @property
    def entity_json(self):
        """
        Gets the entity_json of this BatchErrorMessage.

        :return: The entity_json of this BatchErrorMessage.
        :rtype: str
        """
        return self._entity_json

    @entity_json.setter
    def entity_json(self, entity_json):
        """
        Sets the entity_json of this BatchErrorMessage.

        :param entity_json: The entity_json of this BatchErrorMessage.
        :type: str
        """
        if entity_json is None:
            raise ValueError("Invalid value for `entity_json`, must not be `None`")

        self._entity_json = entity_json

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        if not isinstance(other, BatchErrorMessage):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



import os
# disable Kensu collector when requested to do so (e.g. inside Apache Spark executor nodes)
if "KSU_DISABLE_PY_COLLECTOR" in os.environ:
    from pickle import *
else:
    import pickle as pk
    from .pickle import wrap_dump,wrap_load
    if hasattr(pk, "dump"):
        dump = wrap_dump(pk.dump)
    if hasattr(pk, "load"):
        load = wrap_load(pk.load)



import os
import logging

# This function takes the fully classified object name, say: <package>.<name>.
# Returns the static object instance on the heap
def ref_scala_object(jvm, object_name):
    reflection = False
    if reflection:
      clazz = jvm.java.lang.Class.forName(object_name + "$")
      ff = clazz.getDeclaredField("MODULE$")
      o = ff.get(None)
      return o
    else:
      # it seems in paid databricks (sometimes) reflection do not work...
      class_name_parts = (object_name + "$.MODULE$").split(".")
      o = jvm
      for p in class_name_parts:
        o = getattr(o, p)
      return o

def get_conf_path(self, default = "conf.ini"):
    return os.environ["CONF_FILE"] if "CONF_FILE" in os.environ else default

def wait_for_dam_completion(spark, dam_wait_timeout_secs):
    sc = spark.sparkContext
    jvm = sc._jvm
    # call DamClientActorSystem.destroyActorSystem(actorSysTimeout: Duration = actorSysTimeout, futuresTimeout: Duration = futuresTimeout)
    dam_client_class = ref_scala_object(jvm, "io.kensu.third.integration.spline.DamClientActorSystem")
    duration_class = ref_scala_object(jvm, "scala.concurrent.duration.Duration")
    duration_30s = duration_class.apply(30, "second")
    duration_wait_timeout = duration_class.apply(dam_wait_timeout_secs, "second")
    return dam_client_class.destroyActorSystem(duration_30s, duration_wait_timeout)


def patched_spark_stop(wrapped, dam_wait_timeout_secs):
    def wrapper(self):
        try:
            logging.info('DAM: in spark.stop, waiting for DAM reporting to finish')
            wait_for_dam_completion(self, dam_wait_timeout_secs)
            logging.info('DAM: in spark.stop, waiting for DAM done, stopping spark context')
        except:
            import traceback
            logging.info("unexpected DAM issue: {}".format(traceback.format_exc()))
        result = wrapped(self)
        logging.info('DAM: in spark.stop, spark context stopped')
        return result

    wrapper.__doc__ = wrapped.__doc__
    return wrapper


def rand_str():
    import random
    import string
    return ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(8))


def ensure_dir_exists(path):
    import os
    import errno
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

def patched_spark_createDataFrame_pandas(wrapped, pandas_to_spark_df_via_tmp_file, tmp_dir=None):
    def wrapper(self, data, *args, **kwargs):
        try:
            logging.info('Kensu: before calling spark.createDataFrame')
            from kensu import pandas as ksu_pandas
            if isinstance(data, ksu_pandas.data_frame.DataFrame):
                logging.info('Kensu: replacing kensu.pandas.data_frame.DataFrame wrapper into a original pandas...')
                if not pandas_to_spark_df_via_tmp_file:
                    #  just unwrap the Kensu wrapper, so spark at least works
                    data = data.get_df() # type: pandas.DataFrame
                    # to ensure old-style spark.createDataFrame works
                    # even if passed extra kwarg when KSU tmp file is disabled
                    # we delete the key ksu_tmp_file_prefix if it exists
                    kwargs.pop('ksu_tmp_file_prefix', None)
                else:
                    # given `spark_df=spark.createDataFrame(pandas_df)`, it would be replaced into:
                    # res.to_parquet('/tmp/pandas_to_spark/some_tmp_file.parquet', index=False)
                    # spark_df = spark.read.parquet('file:/tmp/pandas_to_spark/some_tmp_file.parquet')
                    tmp_file_prefix = kwargs.get('ksu_tmp_file_prefix') or 'pandasdf_to_spark'
                    tmp_file_name = tmp_file_prefix + '-' + rand_str()
                    import os
                    # create a unique dir with unique file, so both LDS and file are unique - thus do not clash in Kensu
                    final_tmp_dir = os.path.join(tmp_dir, tmp_file_name)
                    ensure_dir_exists(final_tmp_dir)
                    tmp_file_path = os.path.join(final_tmp_dir, tmp_file_name + '.parquet')
                    # write to a temp file locally with pandas, using Kensu wrapper to capture the lineage
                    abs_tmp_file_name = os.path.abspath(tmp_file_path)
                    logging.info('Kensu: spark.createDataFrame(pandas_df): Kensu will write pandas_df to a temp file - '+ abs_tmp_file_name)
                    data.to_parquet(abs_tmp_file_name, index=False)
                    # here do not call the Spark's original implementation of spark.createDataFrame(),
                    # and just read tmp file with spark, so we'd have the lineage
                    spark_temp_file_path = 'file:' + abs_tmp_file_name
                    logging.info('Kensu: spark.createDataFrame(pandas_df): Kensu will read to a temp file in spark - ' + spark_temp_file_path)
                    spark_df = self.read.parquet(spark_temp_file_path)
                    return spark_df
            logging.info('Kensu: done modifying arguments to spark.createDataFrame')
        except:
            import traceback
            logging.info("unexpected DAM issue: {}".format(traceback.format_exc()))
        result = wrapped(self, data, *args, **kwargs)
        logging.info('Kensu: end of spark.createDataFrame')
        return result

    wrapper.__doc__ = wrapped.__doc__
    return wrapper


def patched_dataframe_write(wrapped):
    def wrapper(self):
        try:
            logging.info('DAM: in patched DataFrame.write, marking DataFrame as cached')
            from pyspark.sql import DataFrameWriter
            result = DataFrameWriter(self.cache())
            logging.info('DAM: in patched DataFrame.write, returning result')
        except:
            import traceback
            logging.info("DAM: unexpected issue in DataFrame.write: {}".format(traceback.format_exc()))
            from pyspark.sql import DataFrameWriter
            result = DataFrameWriter(self.cache())
        return result

    return property(wrapper)


def patched_dataframe_toPandas(wrapped):
    def wrapper(self):
        try:
            logging.info('DAM: in patched DataFrame.toPandas()')
            result = wrapped(self)
            logging.info('DAM: in patched DataFrame.toPandas(), returning result')
        except:
            import traceback
            logging.info("DAM: unexpected issue in DataFrame.toPandas(): {}".format(traceback.format_exc()))
            result = wrapped(self)
        return result

    return wrapper


def get_jvm_from_df(dataframe, # type: DataFrame
                    ):
    sql_context = dataframe.sql_ctx # type: SQLContext
    spark = sql_context.sparkSession
    return spark.sparkContext._jvm


def call_df_as_dam_datasource(df,  # type: DataFrame
                              location, datasource_type,
                              name=None, # type: str
                              logical_name=None # type: str
                             ):
    jvm = get_jvm_from_df(df)
    w = jvm.io.kensu.dam.lineage.spark.lineage.Implicits.SparkDataFrameDAMWrapper(df._jdf)
    return w.reportAsDamDatasource(location, datasource_type, jvm.scala.Option.apply(name), jvm.scala.Option.apply(logical_name))


def call_df_as_dam_jdbc_datasource(df,  # type: DataFrame
                                  dbType,  # type: str
                                  schemaName, # type: str
                                  tableName, # type: str
                                  maybeDatabaseName, # type: str
                                  name=None, # type: str
                                  logical_name=None # type: str
                                  ):
    jvm = get_jvm_from_df(df)
    w = jvm.io.kensu.dam.lineage.spark.lineage.Implicits.SparkDataFrameDAMWrapper(df._jdf)
    return w.reportAsDamJdbcDatasource(dbType, schemaName, tableName, jvm.scala.Option.apply(maybeDatabaseName), jvm.scala.Option.apply(name), jvm.scala.Option.apply(logical_name))


def do_disable_spark_writes(spark):
    """
    Disables spark writes, but keeps DAM reporting.
    Affects only the writes which use DataFrameWriter - dataFrame.write.someMethod(), including .csv(), .save(), .saveAsTable().

    Note: This function MUST be run before first spark write, otherwise spark writes will not be disabled!
    """
    logging.info('DAM disabling df.write')
    sc = spark.sparkContext
    jvm = sc._jvm
    spark_writes_disabler = ref_scala_object(jvm, "io.kensu.third.integration.spark.nowrites.SparkWritesDisabler")
    spark_writes_disabler.disableSparkWrites()
    logging.info('DAM disabling df.write done')


def set_fake_timestamp(spark, mocked_timestamp):
    """
    Sets a fake timestamp to be reported to DAM
    """
    logging.info('DAM set_fake_timestamp starting')
    sc = spark.sparkContext
    jvm = sc._jvm
    spark_writes_disabler = ref_scala_object(jvm, "io.kensu.third.integration.TimeUtils")
    force_override = False
    spark_writes_disabler.setMockedTime(int(mocked_timestamp), force_override)
    logging.info('DAM set_fake_timestamp done')


def add_ds_path_sanitizer(spark, search_str, replacement_str):
    """
    Sets a fake timestamp to be reported to DAM
    """
    logging.info('DAM add_ds_path_sanitizer starting')
    sc = spark.sparkContext
    jvm = sc._jvm
    spark_writes_disabler = ref_scala_object(jvm, "io.kensu.third.integration.DataSourceConv")
    spark_writes_disabler.addDsPathSanitizer(search_str, replacement_str)
    logging.info('DAM add_ds_path_sanitizer done')

def report_df_as_dam_datasource():
    def report_as_dam_datasource(self, # type: DataFrame
                                 location, # type: str
                                 datasource_type=None, # type: str
                                 name=None, # type: str
                                 logical_name=None # type: str
                                 ):
        try:
            logging.info('DAM: in df.report_as_dam_datasource')
            call_df_as_dam_datasource(self, location, datasource_type or 'in-memory-dataframe', name, logical_name)
            logging.info('DAM: df.report_as_dam_datasource started in background')
        except:
            import traceback
            logging.info("unexpected DAM issue: {}".format(traceback.format_exc()))
        return self
    return report_as_dam_datasource


def report_df_as_dam_jdbc_datasource():
    def report_as_dam_jdbc_datasource(self,  # type: DataFrame
                                      db_type,  # type: str
                                      schema_name,  # type: str
                                      table_name,  # type: str
                                      maybe_db_name=None,  # type: str
                                      name=None, # type: str
                                      logical_name=None # type: str
                                      ):
        try:
            logging.info('DAM: in df.report_as_dam_jdbc_datasource')
            call_df_as_dam_jdbc_datasource(self, db_type, schema_name, table_name, maybe_db_name, name, logical_name)
            logging.info('DAM: df.report_as_dam_jdbc_datasource started in background')
        except:
            import traceback
            logging.info("unexpected DAM issue: {}".format(traceback.format_exc()))
        return self
    return report_as_dam_jdbc_datasource


def report_df_as_kpi():
    def report_as_kpi(self,  # type: DataFrame
                      name,  # type: str
                      logical_name=None,  # type: str
                      datasource_type="Business metric (KPI)" # type: str
                      ):
        try:
            logging.info('DAM: in df.report_as_kpi')
            jvm = get_jvm_from_df(self)
            w = jvm.io.kensu.dam.lineage.spark.lineage.Implicits.SparkDataFrameDAMWrapper(self._jdf)
            w.reportAsKPI(name, logical_name or name, datasource_type)
            logging.info('DAM: df.report_as_kpi started in background')
        except:
            import traceback
            logging.info("unexpected DAM issue: {}".format(traceback.format_exc()))
        return self
    return report_as_kpi


def patch_dam_df_helpers():
    try:
        logging.info('Adding DataFrame.report_as_dam_datasource')
        from pyspark.sql import DataFrame
        DataFrame.report_as_dam_datasource = report_df_as_dam_datasource()
        logging.info('done adding DataFrame.report_as_dam_datasource')
    except:
        import traceback
        logging.info("unexpected issue when patching DataFrame.report_as_dam_datasource: {}".format(traceback.format_exc()))
    try:
        logging.info('Adding DataFrame.report_as_dam_jdbc_datasource')
        from pyspark.sql import DataFrame
        DataFrame.report_as_dam_jdbc_datasource = report_df_as_dam_jdbc_datasource()
        logging.info('done adding DataFrame.report_as_dam_jdbc_datasource')
    except:
        import traceback
        logging.info("unexpected issue when patching DataFrame.report_as_dam_jdbc_datasource: {}".format(traceback.format_exc()))
    try:
        logging.info('Adding DataFrame.report_as_kpi')
        from pyspark.sql import DataFrame
        DataFrame.report_as_kpi = report_df_as_kpi()
        logging.info('done adding DataFrame.report_as_kpi')
    except:
        import traceback
        logging.info("unexpected issue when patching DataFrame.report_as_kpi: {}".format(traceback.format_exc()))


def join_paths(maybe_directory, # type: str
               file # type: str
               ):
    if maybe_directory is not None and maybe_directory.strip():
        import os
        return os.path.join(maybe_directory, file)
    else:
        return file


def convert_naming_rules(rules # type: list[str]
                         ):
    # a rule looks like this: dam-spline-persistence/hive-query-results->>File
    return list(["{}->>{}".format(matcher, formatter) for (matcher, formatter) in rules])


def scala_ds_and_schema_to_py(dam, ds, schema):
    from kensu.utils.dsl.extractors.external_lineage_dtos import KensuDatasourceAndSchema
    from kensu.client import DataSourcePK, DataSource, FieldDef, SchemaPK, Schema
    pl_ref = dam.UNKNOWN_PHYSICAL_LOCATION.to_ref()
    py_ds = DataSource(name=ds.name(),
                       format=ds.format(),
                       categories=list(ds.categories()),
                       pk=DataSourcePK(location=ds.pk().location(), physical_location_ref=pl_ref))

    # schema
    fields = [
        FieldDef(name=f.name(), field_type=f.fieldType(), nullable=f.nullable())
        for f in schema.fields()
    ]
    py_schema = Schema(
        name=schema.name(),
        pk=SchemaPK(py_ds.to_ref(),
                    fields=fields))

    def f_publish_stats(lineage_run_id):
        schema.attachStatsTo(lineage_run_id)

    def f_get_stats():
        # ******************************************************************************************
        # WARNING : Do not return anything else than None, because f_publish_stats is preferred giving the responsibility
        # to Spark to send the data stats only when it's available (when computation inside scala Future is completed)
        # ******************************************************************************************
        return None
    return KensuDatasourceAndSchema(ksu_ds=py_ds, ksu_schema=py_schema, f_get_stats=f_get_stats, f_publish_stats = f_publish_stats)


def j2py_dict_of_lists(jdict):
    r = {}
    for k, jlist in jdict.items():
        r[k] = list(jlist)
    return r


def get_inputs_lineage_fn(dam, df):
    jvm = get_jvm_from_df(df)
    # call def fetchToPandasReport(
    #     df: DataFrame,
    #     name: String = UUID.randomUUID().toString,
    #     logicalName: Option[String] = None,
    #     datasourceType: String = InternalDatasourceTags.TAG_INPUT_LINEAGE_ONLY,
    #     timeout: Duration
    #   )
    dam_client_class = ref_scala_object(jvm, "io.kensu.third.integration.dataconversions.SparkToPandasConvTracker")
    duration_300s =  ref_scala_object(jvm, "scala.concurrent.duration.Duration").apply(300, "second")
    # FIXME: it should NOT wait for datasts -> check
    import uuid
    virtual_ds_name = str(uuid.uuid1())
    no_logical_name = jvm.scala.Option.apply(None)
    lineage_from_scala = dam_client_class.fetchToPandasReport(
        df._jdf,
        virtual_ds_name,
        no_logical_name,
        # ds_type,
        duration_300s
    )
    # logging.info('lineage_from_scala:', lineage_from_scala)
    df.lineage_from_scala = lineage_from_scala
    from kensu.utils.dsl.extractors.external_lineage_dtos import GenericComputedInMemDs, ExtDependencyEntry
    # case class SimpleInputDsSchemaAndLineage(
    #   ds: DataSource,
    #   schema: Schema,
    #   lineage: Map[String, Seq[String]]
    # )
    #
    # case class SimpleInputsLineageReport(
    #   lineage: Seq[SimpleInputDsSchemaAndLineage]
    # )
    lineage_info = [
        ExtDependencyEntry(
            input_ds=scala_ds_and_schema_to_py(dam, e.ds(), e.schema()),
            lineage=j2py_dict_of_lists(e.lineage())
        ) for e in lineage_from_scala.lineage()
    ]
    logging.info('lineage_info:', lineage_info)
    return GenericComputedInMemDs(inputs=list([x.input_ds for x in lineage_info]), lineage=lineage_info)


def kwargs_or_conf_or_default(key, default, arg=None, kw=None, conf=None, tpe=None):
    if arg is not None:
        return arg
    elif key in kw and kw[key] is not None:
        return kw[key]
    elif key in conf and conf.get(key) is not None:
        if default is not None and tpe is None:
            tpe = type(default)
        r = conf.get(key)
        if tpe is list:
            r = r.replace(" ", "").split(",")
        elif tpe is bool:
            r = conf.getboolean(key)
        elif tpe is not None:
            r = tpe(r)
        return r
    else:
        return default

"""
Assuming Datasource name is /a/b/c.ext
Allowed values for parameters:
- datasource_short_name_strategy:
  * None (default) - (file) name of DS if use_short_datasource_names else full_path
  * 'File' - c.ext
  * 'LastTwoFoldersAndFile' - a/b/c.ext
  * 'LastFolderAndFile' - b/c.ext
  * 'LastFolder' - b
  * 'PathBasedRule' - use datasource_short_name_strategy_rules
- logical_datasource_name_strategy:
  * None (default) - do not set logical datasource
  * 'File' - c.ext
  * 'LastTwoFoldersAndFile' - a/b/c.ext
  * 'LastFolderAndFile' - b/c.ext
  * 'LastFolder' - b
  * 'PathBasedRule' - use logical_datasource_name_strategy_rules
- datasource_short_name_strategy_rules or logical_datasource_name_strategy_rules:
  * pass a list of tuples (string_matcher, naming_strategy) like this:
    [("features_with_attrition_horizon_label", "File"),
     ("attri_pro_risk_indicator_disabled_3", "LastTwoFoldersAndFile"),
     # default - empty string matches everything
     ("", "LastFolderAndFile")]
- lineage_missing_fallback_strategy:
    * 'CaseInsensitiveColumnNameMatcherStrat'
    * 'CaseSensitiveColumnNameMatcherStrat'
    * 'AllToAllLineageStrat'
    * 'OutFieldEndsWithInFieldNameLineageStrat'
    * 'OutFieldStartsWithInFieldNameLineageStrat'
"""
def init_kensu_spark(
        spark=None,
        ingestion_url=None,
        ingestion_token=None,
        is_offline=None,
        is_dam_tracking_enabled=True,
        dam_logs_directory=None,
        offline_mode_file_name=None,
        explicit_process_name=None,
        explicit_process_run_name=None,
        explicit_code_repo=None,
        explicit_code_version=None,
        explicit_code_maintainers=None,
        explicit_lauched_by_user=None,
        enable_entity_compaction=None,
        allow_invalid_ssl_certificates=None,
        stats=None,
        input_stats=None,
        input_stats_for_only_used_columns=None,
        input_stats_keep_filters=None,
        input_stats_compute_quantiles=None,
        input_stats_cache_bypath=None,
        input_stats_coalesce_enabled=None,
        input_stats_coalesce_workers=None,
        output_stats_compute_quantiles=None,
        output_stats_cache_bypath=None,
        output_stats_coalesce_enabled=None,
        output_stats_coalesce_workers=None,
        cache_output_for_datastats=None,
        use_short_datasource_names=None,
        dam_shutdown_timeout_secs=None,
        dam_debug_file_level=None,
        dam_debug_file_enabled=None,
        datasource_short_name_strategy=None,
        datasource_short_name_strategy_rules=None,
        logical_datasource_name_strategy=None,
        logical_datasource_name_strategy_rules=None,
        lineage_missing_fallback_strategy=None,
        capture_spark_logs=None,
        h2o=None,
        add_dam_dataframe_helpers=None,
        disable_spark_writes=None,
        environment=None,
        fake_timestamp=None,
        projects=None,  # type: list[str]
        project=None,
        h2o_create_virtual_training_datasource=None,
        pandas_conversions=None,
        pandas_to_spark_df_via_tmp_file=None,
        pandas_to_spark_tmp_dir=None,
        **kwargs
):
    if is_dam_tracking_enabled:
        logging.info("Initializing DAM...")
        jvm = spark.sparkContext._jvm

        from configparser import ConfigParser, ExtendedInterpolation
        config = ConfigParser(interpolation=ExtendedInterpolation())
        conf_path = get_conf_path("conf.ini")
        try:
            config.read(conf_path)
        except:
            logging.warning(f"Cannot load config from file `%s`" % (conf_path))

        kensu_conf = config['kensu'] if config.has_section('kensu') else config['DEFAULT']

        ingestion_url = kwargs_or_conf_or_default('api_url',None,ingestion_url, kw=kwargs, conf=kensu_conf)
        ingestion_token = kwargs_or_conf_or_default('api_token',None,ingestion_token, kw=kwargs, conf=kensu_conf)
        is_offline = kwargs_or_conf_or_default('is_offline',False,is_offline, kw=kwargs, conf=kensu_conf, tpe=bool)
        dam_logs_directory = kwargs_or_conf_or_default('dam_logs_directory',None,dam_logs_directory, kw=kwargs, conf=kensu_conf)
        offline_mode_file_name = kwargs_or_conf_or_default('offline_mode_file_name','dam-offline.log',offline_mode_file_name, kw=kwargs, conf=kensu_conf)
        explicit_process_name = kwargs_or_conf_or_default('process_name',None,explicit_process_name, kw=kwargs, conf=kensu_conf)
        explicit_process_run_name = kwargs_or_conf_or_default('explicit_process_run_name',None,explicit_process_run_name, kw=kwargs, conf=kensu_conf)
        explicit_code_repo = kwargs_or_conf_or_default('explicit_code_repo',None,explicit_code_repo, kw=kwargs, conf=kensu_conf)
        explicit_code_version = kwargs_or_conf_or_default('explicit_code_version',None,explicit_code_version, kw=kwargs, conf=kensu_conf)
        explicit_code_maintainers = kwargs_or_conf_or_default('explicit_code_maintainers',None,explicit_code_maintainers, kw=kwargs, conf=kensu_conf)
        explicit_lauched_by_user = kwargs_or_conf_or_default('explicit_lauched_by_user',None,explicit_lauched_by_user, kw=kwargs, conf=kensu_conf)
        enable_entity_compaction = kwargs_or_conf_or_default('enable_entity_compaction',True,enable_entity_compaction, kw=kwargs, conf=kensu_conf, tpe=bool)
        allow_invalid_ssl_certificates = kwargs_or_conf_or_default('allow_invalid_ssl_certificates',False,allow_invalid_ssl_certificates, kw=kwargs, conf=kensu_conf, tpe=bool)
        stats = kwargs_or_conf_or_default('stats',True,stats, kw=kwargs, conf=kensu_conf, tpe=bool)
        input_stats = kwargs_or_conf_or_default('input_stats',True,input_stats, kw=kwargs, conf=kensu_conf, tpe=bool)
        input_stats_for_only_used_columns = kwargs_or_conf_or_default('input_stats_for_only_used_columns',True,input_stats_for_only_used_columns, kw=kwargs, conf=kensu_conf, tpe=bool)
        input_stats_keep_filters = kwargs_or_conf_or_default('input_stats_keep_filters',True,input_stats_keep_filters, kw=kwargs, conf=kensu_conf, tpe=bool)
        input_stats_compute_quantiles =  kwargs_or_conf_or_default('input_stats_compute_quantiles',False,input_stats_compute_quantiles, kw=kwargs, conf=kensu_conf, tpe=bool)
        input_stats_cache_bypath = kwargs_or_conf_or_default('input_stats_cache_bypath',True,input_stats_cache_bypath, kw=kwargs, conf=kensu_conf, tpe=bool)
        input_stats_coalesce_enabled = kwargs_or_conf_or_default('input_stats_coalesce_enabled',True,input_stats_coalesce_enabled, kw=kwargs, conf=kensu_conf, tpe=bool)
        input_stats_coalesce_workers = kwargs_or_conf_or_default('input_stats_coalesce_workers',1,input_stats_coalesce_workers, kw=kwargs, conf=kensu_conf, tpe=int)
        output_stats_compute_quantiles = kwargs_or_conf_or_default('output_stats_compute_quantiles',False,output_stats_compute_quantiles, kw=kwargs, conf=kensu_conf, tpe=bool)
        output_stats_cache_bypath =  kwargs_or_conf_or_default('output_stats_cache_bypath',False,output_stats_cache_bypath, kw=kwargs, conf=kensu_conf, tpe=bool)
        output_stats_coalesce_enabled = kwargs_or_conf_or_default('output_stats_coalesce_enabled',False,output_stats_coalesce_enabled, kw=kwargs, conf=kensu_conf, tpe=bool)
        output_stats_coalesce_workers = kwargs_or_conf_or_default('output_stats_coalesce_workers',100,output_stats_coalesce_workers, kw=kwargs, conf=kensu_conf, tpe=int)
        cache_output_for_datastats = kwargs_or_conf_or_default('cache_output_for_datastats',False,cache_output_for_datastats, kw=kwargs, conf=kensu_conf, tpe=bool)
        use_short_datasource_names = kwargs_or_conf_or_default('use_short_datasource_names',True,use_short_datasource_names, kw=kwargs, conf=kensu_conf, tpe=bool)
        dam_shutdown_timeout_secs = kwargs_or_conf_or_default('dam_shutdown_timeout_secs',10 * 60,dam_shutdown_timeout_secs, kw=kwargs, conf=kensu_conf, tpe=int)
        dam_debug_file_level =  kwargs_or_conf_or_default('dam_debug_file_level','INFO',dam_debug_file_level, kw=kwargs, conf=kensu_conf)
        dam_debug_file_enabled = kwargs_or_conf_or_default('dam_debug_file_enabled',False,dam_debug_file_enabled, kw=kwargs, conf=kensu_conf, tpe=bool)
        datasource_short_name_strategy = kwargs_or_conf_or_default('datasource_short_name_strategy',None,datasource_short_name_strategy, kw=kwargs, conf=kensu_conf)
        datasource_short_name_strategy_rules = kwargs_or_conf_or_default('datasource_short_name_strategy_rules',None,datasource_short_name_strategy_rules, kw=kwargs, conf=kensu_conf)
        logical_datasource_name_strategy = kwargs_or_conf_or_default('logical_naming',None,logical_datasource_name_strategy, kw=kwargs, conf=kensu_conf)
        logical_datasource_name_strategy_rules = kwargs_or_conf_or_default('logical_datasource_name_strategy_rules',None,logical_datasource_name_strategy_rules, kw=kwargs, conf=kensu_conf)
        lineage_missing_fallback_strategy = kwargs_or_conf_or_default('lineage_missing_fallback_strategy',None,lineage_missing_fallback_strategy, kw=kwargs, conf=kensu_conf)
        capture_spark_logs = kwargs_or_conf_or_default('capture_spark_logs',False,capture_spark_logs, kw=kwargs, conf=kensu_conf, tpe=bool)
        h2o = kwargs_or_conf_or_default('h2o',False,h2o, kw=kwargs, conf=kensu_conf, tpe=bool)
        add_dam_dataframe_helpers = kwargs_or_conf_or_default('add_dam_dataframe_helpers',True,add_dam_dataframe_helpers, kw=kwargs, conf=kensu_conf, tpe=bool)
        disable_spark_writes = kwargs_or_conf_or_default('disable_spark_writes',False,disable_spark_writes, kw=kwargs, conf=kensu_conf, tpe=bool)
        environment =  kwargs_or_conf_or_default('environment',None,environment, kw=kwargs, conf=kensu_conf)
        fake_timestamp =  kwargs_or_conf_or_default('fake_timestamp',None,fake_timestamp, kw=kwargs, conf=kensu_conf,tpe=int)
        projects =  kwargs_or_conf_or_default('projects',None,projects, kw=kwargs, conf=kensu_conf,tpe=list)  # type: list[str]
        project =  kwargs_or_conf_or_default('project_names',None,project, kw=kwargs, conf=kensu_conf)
        h2o_create_virtual_training_datasource = kwargs_or_conf_or_default('h2o_create_virtual_training_datasource',True,h2o_create_virtual_training_datasource, kw=kwargs, conf=kensu_conf, tpe=bool)
        pandas_conversions = kwargs_or_conf_or_default('pandas_conversions',True,pandas_conversions, kw=kwargs, conf=kensu_conf, tpe=bool)
        pandas_to_spark_df_via_tmp_file = kwargs_or_conf_or_default('pandas_to_spark_df_via_tmp_file',True,pandas_to_spark_df_via_tmp_file, kw=kwargs, conf=kensu_conf, tpe=bool)
        pandas_to_spark_tmp_dir = kwargs_or_conf_or_default('pandas_to_spark_tmp_dir','/tmp/spark-to-pandas-tmp',pandas_to_spark_tmp_dir, kw=kwargs, conf=kensu_conf)


        try:
            # Not initializing ML trackers yet
            # injected_classes = jvm.io.kensu.third.integration.spark.model.DamModelPublisher.activate().toString()
            ###  Get notebook name ...
            #### see https://github.com/jupyter/notebook/issues/1000#issuecomment-359875246
            import json
            import os.path
            import re
            import requests
            try:  # Python 3
                from urllib.parse import urljoin
            except ImportError:  # Python 2
                from urlparse import urljoin

            def get_notebook_name():
                """
                Return the full path of the jupyter notebook.
                """
                try:  # Python 3
                    from notebook.notebookapp import list_running_servers
                except ImportError:  # Python 2
                    try:
                        import warnings
                        from IPython.utils.shimmodule import ShimWarning
                        with warnings.catch_warnings():
                            warnings.simplefilter("ignore", category=ShimWarning)
                            from IPython.html.notebookapp import list_running_servers
                    except ImportError:  # Probably pyspark script is run without IPython/Jupyter
                        if not explicit_process_name:
                            logging.warning('WARN Unable to automatically extract Jupyter/pyspark notebook name (did you run it without jupyter?)')
                        return ['', explicit_process_name or 'Unknown pyspark filename']
                try:
                    import ipykernel
                    kernel_id = re.search('kernel-(.*).json',
                                          ipykernel.connect.get_connection_file()).group(1)
                    servers = list_running_servers()
                    for ss in servers:
                        response = requests.get(urljoin(ss['url'], 'api/sessions'),
                                                params={'token': ss.get('token', '')})
                        for nn in json.loads(response.text):
                            if nn['kernel']['id'] == kernel_id:
                                server = ss
                                notebooks_path = server['notebook_dir']
                                return [notebooks_path, nn['notebook']['path']]
                except Exception as e:
                    if not explicit_process_name:
                        logging.warning('WARN Unable to automatically extract pyspark notebook name')
                    return ['', explicit_process_name or 'Unknown pyspark filename']

            notebooks_path, notebook_name = get_notebook_name()
            logging.info("Notebook name " + notebook_name)

            ### Configuration for tracker
            global dam_url
            dam_url = os.environ.get("DAM_INGESTION_URL") or ingestion_url

            damIngestionUrl = jvm.scala.Option.apply(dam_url)
            t2 = jvm.scala.Tuple2
            providerClassName = t2("dam.activity.spark.environnement.provider", "io.kensu.dam.lineage.spark.lineage.DefaultSparkEnvironnementProvider")

            if isinstance(explicit_code_maintainers, str):
                explicit_code_maintainers = [explicit_code_maintainers]
            user_names = explicit_code_maintainers or [os.environ.get("USER", "jupyter-notebook")]
            global dam_user_name   # FIXME: why do we need this global? I think we don't?
            jlist = jvm.java.util.ArrayList()
            for user_name in user_names:
                jlist.add(user_name)
                dam_user_name = user_name
            maintainers = t2("dam.activity.code.maintainers", jlist)
            ### How to get GIT info here?
            #         explicit_code_repo=None,
            #         explicit_code_version=None,
            #         explicit_code_maintainers=None,
            #         explicit_lauched_by_user=None,
            code_repo_name = explicit_code_repo or os.environ.get("DAM_CODE_REPOSITORY", "Jupyter-Notebook:spark-" + spark.version+"::"+notebooks_path)
            repository = t2("dam.activity.code.repository", code_repo_name)
            from datetime import datetime
            d = datetime.now()

            code_version_value = explicit_code_version or os.environ.get("DAM_CODE_VERSION", notebook_name + "::" + str(d))
            version = t2("dam.activity.code.version", code_version_value)

            global dam_version
            dam_version = code_version_value
            user = t2("dam.activity.user", explicit_lauched_by_user or dam_user_name)
            organization = t2("dam.activity.organization", "Unknown")
            properties = jvm.scala.collection.mutable.HashSet()

            def add_prop(name, value):
                if value is not None:
                    properties.add(t2(name, value))

            properties.add(providerClassName)
            properties.add(maintainers)
            properties.add(repository)
            properties.add(version)
            properties.add(user)
            properties.add(organization)
            if ingestion_token:
                properties.add(t2("dam.ingestion.auth.token", ingestion_token))
            if is_offline is not None:
                properties.add(t2("dam.ingestion.is_offline", is_offline))
            if offline_mode_file_name:
                properties.add(t2("dam.ingestion.offline.file", join_paths(dam_logs_directory, offline_mode_file_name)))
            if allow_invalid_ssl_certificates is not None:
                properties.add(t2("dam.ingestion.ignore.ssl.cert", allow_invalid_ssl_certificates))
            if enable_entity_compaction is not None:
                properties.add(t2("dam.ingestion.entity.compaction", enable_entity_compaction))

            if stats is not None:
                properties.add(t2("dam.spark.data_stats.enabled", stats))
            if input_stats is not None:
                properties.add(t2("dam.spark.data_stats.input.enabled", input_stats))
            if input_stats_for_only_used_columns:
                properties.add(t2("dam.spark.data_stats.input.only_used_in_lineage", input_stats_for_only_used_columns))
            if input_stats_keep_filters is not None:
                properties.add(t2("dam.spark.data_stats.input.keep_filters", input_stats_keep_filters))

            add_prop("dam.spark.data_stats.input.computeQuantiles", input_stats_compute_quantiles)
            add_prop("dam.spark.data_stats.input.cachePerPath", input_stats_cache_bypath)
            add_prop("dam.spark.data_stats.input.coalesceEnabled", input_stats_coalesce_enabled)
            add_prop("dam.spark.data_stats.input.coalesceWorkers", input_stats_coalesce_workers)

            add_prop("dam.spark.data_stats.output.computeQuantiles", output_stats_compute_quantiles)
            add_prop("dam.spark.data_stats.output.cachePerPath", output_stats_cache_bypath)
            add_prop("dam.spark.data_stats.output.coalesceEnabled", output_stats_coalesce_enabled)
            add_prop("dam.spark.data_stats.output.coalesceWorkers", output_stats_coalesce_workers)

            if dam_shutdown_timeout_secs is not None:
                properties.add(t2("dam.spark.shutdown_timeout", dam_shutdown_timeout_secs))
            if dam_debug_file_enabled:
                debug_level = dam_debug_file_level or "INFO"
                if explicit_process_name is not None:
                    dam_debug_filename = join_paths(dam_logs_directory,explicit_process_name + ".dam-collector-debug.log")
                else:
                    notebook_file_name = notebook_name.split('/')[-1].split('\\')[-1]  # remove path from notebook name
                    dam_debug_filename = join_paths(dam_logs_directory,notebook_file_name + ".dam-collector-debug.log")
                properties.add(t2("dam.spark.file_debug.level", debug_level))
                properties.add(t2("dam.spark.file_debug.file_name", dam_debug_filename))
                properties.add(t2("dam.spark.file_debug.capture_spark_logs", capture_spark_logs))
                logging.info("Will write dam DAM log to file:" + dam_debug_filename)
            if use_short_datasource_names is not None:
                properties.add(t2("dam.datasources.short.name", use_short_datasource_names))

            if datasource_short_name_strategy_rules is not None:
                datasource_short_name_strategy = 'PathBasedRule'
                converted_rules = convert_naming_rules(datasource_short_name_strategy_rules)
                properties.add(t2("dam.datasources.path_rules.short.naming_strategy", converted_rules))
            if logical_datasource_name_strategy_rules is not None:
                logical_datasource_name_strategy = 'PathBasedRule'
                converted_rules = convert_naming_rules(logical_datasource_name_strategy_rules)
                properties.add(t2("dam.logical.datasources.path_rules.naming_strategy", converted_rules))

            if lineage_missing_fallback_strategy is not None:
                properties.add(t2('dam.spark.lineage.column_lineage_fallback_strategy', lineage_missing_fallback_strategy))

            if datasource_short_name_strategy is not None:
                properties.add(t2("dam.datasources.short.naming_strategy", datasource_short_name_strategy))
            if logical_datasource_name_strategy is not None:
                properties.add(t2("dam.logical.datasources.naming_strategy", logical_datasource_name_strategy))
            if environment is not None:
                properties.add(t2("dam.activity.environment", environment))
            if projects is not None:
                properties.add(t2("dam.activity.projects", projects))
            elif project is not None:
                properties.add(t2("dam.activity.projects", [project]))

            process_name = None
            if explicit_process_name is not None:
                process_name = explicit_process_name
            elif notebook_name != 'Unknown pyspark filename':
                process_name = notebook_name
            if process_name is not None:
                properties.add(t2("dam.activity.explicit.process.name", process_name))

            if explicit_process_run_name is not None:
                properties.add(t2("dam.activity.explicit.process_run.name", process_name))

            maybe_fake_timestamp = fake_timestamp or os.environ.get('DAM_OVERRIDE_TIMESTAMP')
            if maybe_fake_timestamp is not None:
                set_fake_timestamp(spark, maybe_fake_timestamp)
            maybe_ds_path_sanitizer_search = os.environ.get('DAM_DS_PATH_SANITIZER_SEARCH')
            maybe_ds_path_sanitizer_replace = os.environ.get('DAM_DS_PATH_SANITIZER_REPLACE')
            if (maybe_ds_path_sanitizer_search is not None) and (maybe_ds_path_sanitizer_replace is not None):
                add_ds_path_sanitizer(spark, maybe_ds_path_sanitizer_search, maybe_ds_path_sanitizer_replace)

            w = jvm.io.kensu.dam.lineage.spark.lineage.Implicits.SparkSessionDAMWrapper(spark._jsparkSession)
            w.track(damIngestionUrl, jvm.scala.Option.empty(), properties.toSeq())

            if (dam_shutdown_timeout_secs is not None) and (dam_shutdown_timeout_secs > 0):
                logging.info('patching spark.stop to wait for DAM reporting to finish')
                from pyspark.sql import SparkSession
                SparkSession.stop = patched_spark_stop(SparkSession.stop, dam_shutdown_timeout_secs)
                logging.info('patching spark.stop done')

            if disable_spark_writes:
                try:
                    do_disable_spark_writes(spark)
                except:
                    import traceback
                    logging.info("unexpected issue when disabling df.write {}".format(traceback.format_exc()))

            if cache_output_for_datastats and stats:
                try:
                    logging.info('patching DataFrame.write to cache/persist the results to speedup data-stats computation')
                    from pyspark.sql import DataFrame
                    DataFrame.write = patched_dataframe_write(DataFrame.write)
                    logging.info('done patching DataFrame.write')
                except:
                    import traceback
                    logging.info("unexpected issue when patching DataFrame.write: {}".format(traceback.format_exc()))

            if add_dam_dataframe_helpers:
                patch_dam_df_helpers()

            if pandas_conversions:
                try:
                    logging.info('patching DataFrame.toPandas')
                    from pyspark.sql import DataFrame
                    import kensu.pandas as pd
                    DataFrame.toPandas = pd.data_frame.wrap_external_to_pandas_transformation(
                        DataFrame.toPandas,
                        get_inputs_lineage_fn
                    )
                    logging.info('done patching DataFrame.toPandas')
                except:
                    import traceback
                    logging.info("unexpected issue when patching DataFrame.toPandas: {}".format(traceback.format_exc()))
                try:
                    logging.info('patching spark.createDataFrame to work with pandas dataframes')
                    from pyspark.sql import SparkSession
                    SparkSession.createDataFrame = patched_spark_createDataFrame_pandas(
                        SparkSession.createDataFrame,
                        pandas_to_spark_df_via_tmp_file=pandas_to_spark_df_via_tmp_file,
                        tmp_dir=pandas_to_spark_tmp_dir
                    )
                    logging.info('patching  spark.createDataFrame done')
                except:
                    import traceback
                    logging.info("unexpected issue when patching DataFrame.toPandas: {}".format(traceback.format_exc()))
                try:
                    logging.info('adding spark env var spark.executorEnv.KSU_DISABLE_PY_COLLECTOR=true'
                          ' to disable kensu-py collector on executor nodes')
                    spark.conf.set("spark.executorEnv.KSU_DISABLE_PY_COLLECTOR", 'true')
                    spark.sparkContext.environment['KSU_DISABLE_PY_COLLECTOR'] = 'true'
                    logging.info('done adding spark.executorEnv.KSU_DISABLE_PY_COLLECTOR')
                except:
                    import traceback
                    logging.info("unexpected issue when adding spark.executorEnv.KSU_DISABLE_PY_COLLECTOR: {}".format(traceback.format_exc()))


            if 'kensu_py_client' in kwargs:
                if kwargs['kensu_py_client']:
                    from kensu.utils.kensu_provider import KensuProvider as K
                    K().initKensu(do_report=False)

            if h2o:
                # needed to support both ways of importing: from local file or dam-client-python package
                try:
                    from kensu_dam.h2o import dam_patch_h2o
                except ImportError:
                    from dam_h2o_collector import dam_patch_h2o

                if dam_logs_directory is not None:
                    dam_patch_h2o(log_file_prefix=explicit_process_name or '',
                                  debug_file_enabled=True,
                                  debug_stdout_enabled=True,
                                  dam_logs_directory=dam_logs_directory,
                                  create_virtual_training_datasource=h2o_create_virtual_training_datasource)
                else:
                    dam_patch_h2o(debug_file_enabled=False,
                                  debug_stdout_enabled=True,
                                  create_virtual_training_datasource=h2o_create_virtual_training_datasource)


        except Exception as e:
            logging.info("Error when initializing DAM tracker:")
            import traceback
            logging.info(traceback.format_exc())



from psycopg2 import extras


def pg_result_to_dict(cur, res):
    rows=[]
    for row in res:
        row_dict = {}
        idx = 0
        for col in cur.description:
            row_dict.update({str(col.name): row[idx]})
            idx += 1
        rows.append(row_dict)
    return rows


def pg_query_as_dicts(cur, q):
    cur.execute(q)
    res = cur.fetchall()
    return pg_result_to_dict(cur, res)


def get_current_db_info(cur):
    res=pg_query_as_dicts(cur, q='SELECT current_database() AS curr_catalog, current_schema() AS curr_schema')
    for r in res:
        return r.get('curr_catalog'), r.get('curr_schema')
    return None, None

def get_table_schema(cur, table_name):
    # returns [{'field_name': 'start_date', 'field_type': 'date'}]

    cur.execute("select * from {} where false".format(table_name))
    fields_with_typeid = [[col.name, col.type_code] for col in cur.description]

    # below it generates query like this
    # SELECT q.field_name, q.field_type_oid, format_type(q.field_type_oid, NULL)
    # FROM (VALUES ('int', 20), ('dt', 1082)) AS q(field_name, field_type_oid);
    q = """SELECT q.field_name, format_type(q.field_type_oid, NULL) AS field_type
            FROM (VALUES %s) AS q(field_name, field_type_oid)"""
    result = extras.execute_values(cur, q, fields_with_typeid, fetch=True)
    res_dict = pg_result_to_dict(cur=cur, res=result)
    return res_dict



# coding: utf-8

"""
    

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)

    OpenAPI spec version: beta
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


from pprint import pformat

from six import iteritems


class LineageRunPK(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """


    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'process_run_ref': 'ProcessRunRef',
        'lineage_ref': 'ProcessLineageRef',
        'timestamp': 'int'
    }

    attribute_map = {
        'process_run_ref': 'processRunRef',
        'lineage_ref': 'lineageRef',
        'timestamp': 'timestamp'
    }

    def __init__(self, process_run_ref=None, lineage_ref=None, timestamp=None):
        """
        LineageRunPK - a model defined in Swagger
        """

        self._process_run_ref = None
        self._lineage_ref = None
        self._timestamp = None

        self.process_run_ref = process_run_ref
        self.lineage_ref = lineage_ref
        self.timestamp = timestamp

    @property
    def process_run_ref(self):
        """
        Gets the process_run_ref of this LineageRunPK.
        Reference to a process run/execution where this lineage was discovered (e.g. `credit_score_job_123_12`)

        :return: The process_run_ref of this LineageRunPK.
        :rtype: ProcessRunRef
        """
        return self._process_run_ref

    @process_run_ref.setter
    def process_run_ref(self, process_run_ref):
        """
        Sets the process_run_ref of this LineageRunPK.
        Reference to a process run/execution where this lineage was discovered (e.g. `credit_score_job_123_12`)

        :param process_run_ref: The process_run_ref of this LineageRunPK.
        :type: ProcessRunRef
        """
        if process_run_ref is None:
            raise ValueError("Invalid value for `process_run_ref`, must not be `None`")

        self._process_run_ref = process_run_ref

    @property
    def lineage_ref(self):
        """
        Gets the lineage_ref of this LineageRunPK.

        :return: The lineage_ref of this LineageRunPK.
        :rtype: ProcessLineageRef
        """
        return self._lineage_ref

    @lineage_ref.setter
    def lineage_ref(self, lineage_ref):
        """
        Sets the lineage_ref of this LineageRunPK.

        :param lineage_ref: The lineage_ref of this LineageRunPK.
        :type: ProcessLineageRef
        """
        if lineage_ref is None:
            raise ValueError("Invalid value for `lineage_ref`, must not be `None`")

        self._lineage_ref = lineage_ref

    @property
    def timestamp(self):
        """
        Gets the timestamp of this LineageRunPK.
        The timestamp at which the lineage was run

        :return: The timestamp of this LineageRunPK.
        :rtype: int
        """
        return self._timestamp

    @timestamp.setter
    def timestamp(self, timestamp):
        """
        Sets the timestamp of this LineageRunPK.
        The timestamp at which the lineage was run

        :param timestamp: The timestamp of this LineageRunPK.
        :type: int
        """
        if timestamp is None:
            raise ValueError("Invalid value for `timestamp`, must not be `None`")

        self._timestamp = timestamp

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        if not isinstance(other, LineageRunPK):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other
